"""Maxima Hunting dimensionality reduction and related methods."""
from __future__ import annotations

from typing import Callable, Optional, Union

import numpy as np
import scipy.signal
import sklearn.base
import sklearn.utils

from dcor import u_distance_correlation_sqr

from ...._utils import _compute_dependence, _DependenceMeasure
from ....representation import FDataGrid
from ....representation._typing import NDArrayFloat, NDArrayInt

_LocalMaximaSelector = Callable[[FDataGrid], NDArrayInt]


def select_local_maxima(X: FDataGrid, *, order: int = 1) -> NDArrayInt:
    r"""
    Compute local maxima of an array.

    Points near the boundary are considered maxima looking only at one side.

    For flat regions only the boundary points of the flat region could be
    considered maxima.

    Parameters:
        X: Where to compute the local maxima.
        order: How many points on each side to look, to check if
            a point is a maximum in that interval.

    Returns:
        Indexes of the local maxima.

    Examples:
        >>> from skfda import FDataGrid
        >>> from skfda.preprocessing.dim_reduction.variable_selection.\
        ...     maxima_hunting import select_local_maxima
        >>> import numpy as np

        >>> x = FDataGrid(np.array([2, 1, 1, 1, 2, 3, 3, 3, 2, 3, 4, 3, 2]))
        >>> select_local_maxima(x).astype(np.int_)
        array([ 0,  5,  7, 10])

        The ``order`` parameter can be used to check a larger interval to see
        if a point is still a maxima, effectively eliminating small local
        maxima.

        >>> x = FDataGrid(np.array([2, 1, 1, 1, 2, 3, 3, 3, 2, 3, 4, 3, 2]))
        >>> select_local_maxima(x, order=3).astype(np.int_)
        array([ 0,  5, 10])

    """
    X_array = X.data_matrix[0, ..., 0]

    indexes = scipy.signal.argrelextrema(
        X_array,
        comparator=np.greater_equal,
        order=order,
    )[0]

    # Discard flat
    maxima = X_array[indexes]

    left_points = np.take(X_array, indexes - 1, mode='clip')
    right_points = np.take(X_array, indexes + 1, mode='clip')

    is_not_flat = (maxima > left_points) | (maxima > right_points)

    return indexes[is_not_flat]


class MaximaHunting(
    sklearn.base.BaseEstimator,  # type: ignore
    sklearn.base.TransformerMixin,  # type: ignore
):
    r"""
    Maxima Hunting variable selection.

    This is a filter variable selection method for problems with a target
    variable. It evaluates a dependence measure between each point of the
    function and the target variable, and keeps those points in which this
    dependence is a local maximum.

    Selecting the local maxima serves two purposes. First, it ensures that
    the points that are relevant in isolation are selected, as they must
    maximice their dependence with the target variable. Second, the points
    that are relevant only because they are near a relevant point (and are
    thus highly correlated with it) are NOT selected, as only local maxima
    are selected, minimizing the redundancy of the selected variables.

    For a longer explanation about the method, and comparison with other
    functional variable selection methods, we refer the reader to the
    original article :footcite:`berrendero+cuevas+torrecilla_2016_hunting`.

    Parameters:
        dependence_measure (callable): Dependence measure to use. By default,
            it uses the bias corrected squared distance correlation.
        local_maxima_selector (callable): Function to detect local maxima. The
            default is :func:`select_local_maxima` with ``order`` parameter
            equal to one. The original article used a similar function testing
            different values of ``order``.

    Examples:
        >>> from skfda.preprocessing.dim_reduction import variable_selection
        >>> from skfda.preprocessing.dim_reduction.variable_selection.\
        ...     maxima_hunting import select_local_maxima
        >>> from skfda.datasets import make_gaussian_process
        >>> from functools import partial
        >>> import skfda
        >>> import numpy as np

        We create trajectories from two classes, one with zero mean and the
        other with a peak-like mean. Both have Brownian covariance.

        >>> n_samples = 10000
        >>> n_features = 100
        >>>
        >>> def mean_1(t):
        ...     return (np.abs(t - 0.25)
        ...             - 2 * np.abs(t - 0.5)
        ...             + np.abs(t - 0.75))
        >>>
        >>> X_0 = make_gaussian_process(n_samples=n_samples // 2,
        ...                             n_features=n_features,
        ...                             random_state=0)
        >>> X_1 = make_gaussian_process(n_samples=n_samples // 2,
        ...                             n_features=n_features,
        ...                             mean=mean_1,
        ...                             random_state=1)
        >>> X = skfda.concatenate((X_0, X_1))
        >>>
        >>> y = np.zeros(n_samples)
        >>> y [n_samples // 2:] = 1

        Select the relevant points to distinguish the two classes

        >>> local_maxima_selector = partial(select_local_maxima, order=10)
        >>> mh = variable_selection.MaximaHunting(
        ...            local_maxima_selector=local_maxima_selector)
        >>> _ = mh.fit(X, y)
        >>> point_mask = mh.get_support()
        >>> points = X.grid_points[0][point_mask]
        >>> np.allclose(points, [0.5], rtol=0.1)
        True

        Apply the learned dimensionality reduction

        >>> X_dimred = mh.transform(X)
        >>> len(X.grid_points[0])
        100
        >>> X_dimred.shape
        (10000, 1)

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        dependence_measure: _DependenceMeasure = u_distance_correlation_sqr,
        local_maxima_selector: _LocalMaximaSelector = select_local_maxima,
    ) -> None:
        self.dependence_measure = dependence_measure
        self.local_maxima_selector = local_maxima_selector

    def fit(
        self,
        X: FDataGrid,
        y: Union[NDArrayInt, NDArrayFloat],
    ) -> MaximaHunting:  # noqa: D102

        self.features_shape_ = X.data_matrix.shape[1:]
        self.dependence_ = _compute_dependence(
            X.data_matrix,
            y,
            dependence_measure=self.dependence_measure,
        )

        self.indexes_ = self.local_maxima_selector(
            FDataGrid(
                self.dependence_,
                grid_points=X.grid_points,
            ),
        )

        sorting_indexes = np.argsort(self.dependence_[self.indexes_])[::-1]
        self.sorted_indexes_ = self.indexes_[sorting_indexes]

        return self

    def get_support(self, indices: bool = False) -> NDArrayInt:  # noqa: D102
        if indices:
            return self.indexes_

        mask = np.zeros(self.features_shape_[:-1], dtype=bool)
        mask[self.indexes_] = True
        return mask

    def transform(  # noqa: D102
        self,
        X: FDataGrid,
        y: Optional[Union[NDArrayInt, NDArrayFloat]] = None,
    ) -> NDArrayFloat:

        sklearn.utils.validation.check_is_fitted(self)

        if X.data_matrix.shape[1:] != self.features_shape_:
            raise ValueError(
                "The trajectories have a different number of "
                "points than the ones fitted",
            )

        return X.data_matrix[:, self.sorted_indexes_].reshape(X.n_samples, -1)



"""Basic mathematical functionalities of the package.

Defines the basic mathematical operations for classes defined in this
package. FDataBasis and FDataGrid.

"""
import warnings
from builtins import isinstance
from typing import Any, Callable, Optional, TypeVar, Union, cast

import multimethod
import numpy as np
import scipy.integrate

from .._utils import _same_domain, nquad_vec
from ..representation import FData, FDataBasis, FDataGrid
from ..representation._typing import ArrayLike, DomainRange, NDArrayFloat
from ..representation.basis import Basis

Vector = TypeVar(
    "Vector",
    bound=Union[NDArrayFloat, Basis, Callable[[ArrayLike], NDArrayFloat]],
)


def sqrt(fdatagrid: FDataGrid) -> FDataGrid:
    """Perform a element wise square root operation.

    .. deprecated:: 0.6
        Use :func:`numpy.sqrt` function instead.

    Args:
        fdatagrid: Object to whose elements the square root
            operation is going to be applied.

    Returns:
        FDataGrid object whose elements are the square roots of the original.

    """
    warnings.warn(
        "Function sqrt is deprecated. Use numpy.sqrt with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.sqrt(fdatagrid))


def absolute(fdatagrid: FDataGrid) -> FDataGrid:
    """Get the absolute value of all elements in the FDataGrid object.

    .. deprecated:: 0.6
        Use :func:`numpy.absolute` function instead.

    Args:
        fdatagrid: Object from whose elements the absolute value
            is going to be retrieved.

    Returns:
        FDataGrid object whose elements are the absolute values of the
            original.

    """
    warnings.warn(
        "Function absolute is deprecated. Use numpy.absolute with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.absolute(fdatagrid))


def round(  # noqa: WPS125
    fdatagrid: FDataGrid,
    decimals: int = 0,
) -> FDataGrid:
    """Round all elements of the object.

    .. deprecated:: 0.6
        Use :func:`numpy.round` function instead.

    Args:
        fdatagrid: Object to whose elements are going to be
            rounded.
        decimals: Number of decimals wanted. Defaults to 0.

    Returns:
        FDataGrid object whose elements are rounded.

    """
    warnings.warn(
        "Function round is deprecated. Use numpy.round with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.round(fdatagrid, decimals))


def exp(fdatagrid: FDataGrid) -> FDataGrid:
    """Perform a element wise exponential operation.

    .. deprecated:: 0.6
        Use :func:`numpy.exp` function instead.

    Args:
        fdatagrid: Object to whose elements the exponential
            operation is going to be applied.

    Returns:
        FDataGrid object whose elements are the result of exponentiating
            the elements of the original.

    """
    warnings.warn(
        "Function exp is deprecated. Use numpy.exp with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.exp(fdatagrid))


def log(fdatagrid: FDataGrid) -> FDataGrid:
    """Perform a element wise logarithm operation.

    .. deprecated:: 0.6
        Use :func:`numpy.log` function instead.

    Args:
        fdatagrid: Object to whose elements the logarithm
            operation is going to be applied.

    Returns:
        FDataGrid object whose elements are the logarithm of the original.

    """
    warnings.warn(
        "Function log is deprecated. Use numpy.log with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.log(fdatagrid))


def log10(fdatagrid: FDataGrid) -> FDataGrid:
    """Perform an element wise base 10 logarithm operation.

    .. deprecated:: 0.6
        Use :func:`numpy.log10` function instead.

    Args:
        fdatagrid: Object to whose elements the base 10 logarithm
            operation is going to be applied.

    Returns:
        FDataGrid object whose elements are the base 10 logarithm of the
            original.

    """
    warnings.warn(
        "Function log10 is deprecated. Use numpy.log10 with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.log10(fdatagrid))


def log2(fdatagrid: FDataGrid) -> FDataGrid:
    """Perform an element wise binary logarithm operation.

    .. deprecated:: 0.6
        Use :func:`numpy.log2` function instead.

    Args:
        fdatagrid: Object to whose elements the binary logarithm
            operation is going to be applied.

    Returns:
        FDataGrid object whose elements are the binary logarithm of the
            original.

    """
    warnings.warn(
        "Function log2 is deprecated. Use numpy.log2 with a FDataGrid "
        "parameter instead.",
        DeprecationWarning,
    )

    return cast(FDataGrid, np.log2(fdatagrid))


def cumsum(fdatagrid: FDataGrid) -> FDataGrid:
    """Return the cumulative sum of the samples.

    Args:
        fdatagrid (FDataGrid): Object over whose samples the cumulative sum is
            going to be calculated.

    Returns:
        FDataGrid: Object with the sample wise cumulative sum.

    """
    return fdatagrid.copy(
        data_matrix=np.cumsum(fdatagrid.data_matrix, axis=0),
    )


@multimethod.multidispatch
def inner_product(
    arg1: Vector,
    arg2: Vector,
    *,
    _matrix: bool = False,
    _domain_range: Optional[DomainRange] = None,
    **kwargs: Any,
) -> NDArrayFloat:
    r"""
    Return the usual (:math:`L_2`) inner product.

    Calculates the inner product between matching samples in two
    FDataGrid objects.

    For two samples x and y the inner product is defined as:

    .. math::
        \langle x, y \rangle = \sum_i x_i y_i

    for multivariate data and

    .. math::
        \langle x, y \rangle = \int_a^b x(t)y(t)dt

    for functional data.

    The two arguments must have the same number of samples, or one should
    contain only one sample (and will be broadcasted).

    Args:
        arg1: First sample.
        arg2: Second sample.

    Returns:
        Vector with the inner products of each pair of samples.

    Examples:
        This function can compute the multivariate inner product.

        >>> import numpy as np
        >>> from skfda.misc import inner_product
        >>>
        >>> array1 = np.array([1, 2, 3])
        >>> array2 = np.array([4, 5, 6])
        >>> inner_product(array1, array2)
        32

        If the arrays contain more than one sample

        >>> array1 = np.array([[1, 2, 3], [2, 3, 4]])
        >>> array2 = np.array([[4, 5, 6], [1, 1, 1]])
        >>> inner_product(array1, array2)
        array([32, 9])

        The inner product of the :math:'f(x) = x` and the constant
        :math:`y=1` defined over the interval [0,1] is the area of the
        triangle delimited by the the lines y = 0, x = 1 and y = x; 0.5.

        >>> import skfda
        >>>
        >>> x = np.linspace(0,1,1000)
        >>>
        >>> fd1 = skfda.FDataGrid(x,x)
        >>> fd2 = skfda.FDataGrid(np.ones(len(x)),x)
        >>> inner_product(fd1, fd2)
        array([ 0.5])

        If the FDataGrid object contains more than one sample

        >>> fd1 = skfda.FDataGrid([x, np.ones(len(x))], x)
        >>> fd2 = skfda.FDataGrid([np.ones(len(x)), x] ,x)
        >>> inner_product(fd1, fd2).round(2)
        array([ 0.5, 0.5])

        If one argument contains only one sample it is
        broadcasted.

        >>> fd1 = skfda.FDataGrid([x, np.ones(len(x))], x)
        >>> fd2 = skfda.FDataGrid([np.ones(len(x))] ,x)
        >>> inner_product(fd1, fd2).round(2)
        array([ 0.5, 1. ])

        It also work with basis objects

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [0, 1, 0])
        >>> fd2 = skfda.FDataBasis(basis, [1, 0, 0])
        >>> inner_product(fd1, fd2)
        array([ 0.5])

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [[0, 1, 0], [0, 0, 1]])
        >>> fd2 = skfda.FDataBasis(basis, [1, 0, 0])
        >>> inner_product(fd1, fd2)
        array([ 0.5       , 0.33333333])

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [[0, 1, 0], [0, 0, 1]])
        >>> fd2 = skfda.FDataBasis(basis, [[1, 0, 0], [0, 1, 0]])
        >>> inner_product(fd1, fd2)
        array([ 0.5 , 0.25])

    """
    if callable(arg1) and callable(arg2):
        return _inner_product_integrate(
            arg1,
            arg2,
            _matrix=_matrix,
            _domain_range=_domain_range,
        )

    return (
        np.einsum('n...,m...->nm...', arg1, arg2).sum(axis=-1)
        if _matrix else (arg1 * arg2).sum(axis=-1)
    )


@inner_product.register
def _inner_product_fdatagrid(
    arg1: FDataGrid,
    arg2: FDataGrid,
    *,
    _matrix: bool = False,
) -> NDArrayFloat:

    if not np.array_equal(
        arg1.grid_points,
        arg2.grid_points,
    ):
        raise ValueError("Sample points for both objects must be equal")

    d1 = arg1.data_matrix
    d2 = arg2.data_matrix

    einsum_broadcast_list = (np.arange(d1.ndim - 1) + 2).tolist()

    if _matrix:

        d1 = np.copy(d1)

        # Perform quadrature inside the einsum
        for i, s in enumerate(arg1.grid_points[::-1]):
            identity = np.eye(len(s))
            weights = scipy.integrate.simps(identity, x=s)
            index = (slice(None),) + (np.newaxis,) * (i + 1)
            d1 *= weights[index]

        return np.einsum(
            d1,
            [0] + einsum_broadcast_list,
            d2,
            [1] + einsum_broadcast_list,
            [0, 1],
        )

    integrand = d1 * d2

    for g in arg1.grid_points[::-1]:
        integrand = scipy.integrate.simps(
            integrand,
            x=g,
            axis=-2,
        )

    return np.sum(integrand, axis=-1)


@inner_product.register(FDataBasis, FDataBasis)
@inner_product.register(FDataBasis, Basis)
@inner_product.register(Basis, FDataBasis)
@inner_product.register(Basis, Basis)
def _inner_product_fdatabasis(
    arg1: Union[FDataBasis, Basis],
    arg2: Union[FDataBasis, Basis],
    *,
    _matrix: bool = False,
    inner_product_matrix: Optional[NDArrayFloat] = None,
    force_numerical: bool = False,
) -> NDArrayFloat:

    if not _same_domain(arg1, arg2):
        raise ValueError("Both Objects should have the same domain_range")

    if isinstance(arg1, Basis):
        arg1 = arg1.to_basis()

    if isinstance(arg2, Basis):
        arg2 = arg2.to_basis()

    # Now several cases where computing the matrix is preferrable
    #
    # First, if force_numerical is True, the matrix is NOT used
    # Otherwise, if the matrix is given, it is used
    # Two other cases follow

    # The basis is the same: most basis can optimize this case,
    # and also the Gram matrix is cached the first time, so computing
    # it is usually worthwhile
    same_basis = arg1.basis == arg2.basis

    # The number of operations is less using the matrix
    n_ops_best_with_matrix = max(
        arg1.n_samples,
        arg2.n_samples,
    ) > arg1.n_basis * arg2.n_basis

    if not force_numerical and (
        inner_product_matrix is not None
        or same_basis
        or n_ops_best_with_matrix
    ):

        if inner_product_matrix is None:
            inner_product_matrix = arg1.basis.inner_product_matrix(arg2.basis)

        coef1 = arg1.coefficients
        coef2 = arg2.coefficients

        if _matrix:
            return np.einsum(
                'nb,bc,mc->nm',
                coef1,
                inner_product_matrix,
                coef2,
            )

        return (
            coef1
            @ inner_product_matrix
            * coef2
        ).sum(axis=-1)

    return _inner_product_integrate(arg1, arg2, _matrix=_matrix)


def _inner_product_integrate(
    arg1: Callable[[ArrayLike], NDArrayFloat],
    arg2: Callable[[ArrayLike], NDArrayFloat],
    *,
    _matrix: bool = False,
    _domain_range: Optional[DomainRange] = None,
) -> NDArrayFloat:

    domain_range: DomainRange

    if isinstance(arg1, FData) and isinstance(arg2, FData):
        if not np.array_equal(
            arg1.domain_range,
            arg2.domain_range,
        ):
            raise ValueError("Domain range for both objects must be equal")

        domain_range = arg1.domain_range
        len_arg1 = len(arg1)
        len_arg2 = len(arg2)
    else:
        # If the arguments are callables, we need to pass the domain range
        # explicitly. This is used internally for computing the gramian
        # matrix of operators.
        assert _domain_range is not None
        domain_range = _domain_range
        left_domain = np.array(domain_range)[:, 0]
        len_arg1 = len(arg1(left_domain))
        len_arg2 = len(arg2(left_domain))

    def integrand(*args: NDArrayFloat) -> NDArrayFloat:  # noqa: WPS430
        f1 = arg1(args)[:, 0, :]
        f2 = arg2(args)[:, 0, :]

        if _matrix:
            ret = np.einsum('n...,m...->nm...', f1, f2)
            return ret.reshape((-1,) + ret.shape[2:])

        return f1 * f2

    integral = nquad_vec(
        integrand,
        domain_range,
    )

    summation = np.sum(integral, axis=-1)

    if _matrix:
        summation = summation.reshape((len_arg1, len_arg2))

    return summation


def inner_product_matrix(
    arg1: Vector,
    arg2: Optional[Vector] = None,
    **kwargs: Any,
) -> NDArrayFloat:
    """
    Return the inner product matrix between is arguments.

    Args:
        arg1: First sample.
        arg2: Second sample. If it is ``None`` returns the inner product
            between the samples in ``arg1``.
        kwargs: Keyword arguments for inner product.

    Returns:
        Inner product matrix between samples.

    See also:
        :func:`inner_product`

    """
    if isinstance(arg1, Basis):
        arg1 = arg1.to_basis()
    if isinstance(arg2, Basis):
        arg2 = arg2.to_basis()

    if arg2 is None:
        arg2 = arg1

    return inner_product(arg1, arg2, _matrix=True, **kwargs)


def _clip_cosine(array: NDArrayFloat) -> NDArrayFloat:
    """Clip cosine values to prevent numerical errors."""
    small_val = 1e-6

    # If the difference is too large, there could be a problem
    assert np.all((-1 - small_val < array) & (array < 1 + small_val))

    return np.clip(array, -1, 1)


def cosine_similarity(
    arg1: Vector,
    arg2: Vector,
) -> NDArrayFloat:
    r"""
    Return the cosine similarity.

    Calculates the cosine similarity between matching samples in two
    FDataGrid objects.

    For two samples x and y the cosine similarity is defined as:

    .. math::
        \cos \text{sim}(x, y) = \frac{\langle x, y \rangle}{
        \sqrt{\langle x, x \rangle \langle y, y \rangle}}

    where :math:`\langle {}\cdot{}, {}\cdot{} \rangle` is the inner product.

    The two arguments must have the same number of samples, or one should
    contain only one sample (and will be broadcasted).

    Args:
        arg1: First sample.
        arg2: Second sample.

    Returns:
        Vector with the cosine similarity of each pair of samples.

    Examples:
        This function can compute the multivariate cosine similarity.

        >>> import numpy as np
        >>> from skfda.misc import cosine_similarity
        >>>
        >>> array1 = np.array([1, 2, 3])
        >>> array2 = np.array([4, 5, 6])
        >>> cosine_similarity(array1, array2)
        0.9746318461970762

        If the arrays contain more than one sample

        >>> array1 = np.array([[1, 2, 3], [2, 3, 4]])
        >>> array2 = np.array([[4, 5, 6], [1, 1, 1]])
        >>> cosine_similarity(array1, array2)
        array([ 0.97463185,  0.96490128])

        The cosine similarity of the :math:'f(x) = x` and the constant
        :math:`y=1` defined over the interval [0,1] is the area of the
        triangle delimited by the the lines y = 0, x = 1 and y = x; 0.5,
        multiplied by :math:`\sqrt{3}`.

        >>> import skfda
        >>>
        >>> x = np.linspace(0,1,1000)
        >>>
        >>> fd1 = skfda.FDataGrid(x,x)
        >>> fd2 = skfda.FDataGrid(np.ones(len(x)),x)
        >>> cosine_similarity(fd1, fd2)
        array([ 0.8660254])

        If the FDataGrid object contains more than one sample

        >>> fd1 = skfda.FDataGrid([x, np.ones(len(x))], x)
        >>> fd2 = skfda.FDataGrid([np.ones(len(x)), x] ,x)
        >>> cosine_similarity(fd1, fd2).round(2)
        array([ 0.87,  0.87])

        If one argument contains only one sample it is
        broadcasted.

        >>> fd1 = skfda.FDataGrid([x, np.ones(len(x))], x)
        >>> fd2 = skfda.FDataGrid([np.ones(len(x))] ,x)
        >>> cosine_similarity(fd1, fd2).round(2)
        array([ 0.87,  1.  ])

        It also work with basis objects

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [0, 1, 0])
        >>> fd2 = skfda.FDataBasis(basis, [1, 0, 0])
        >>> cosine_similarity(fd1, fd2)
        array([ 0.8660254])

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [[0, 1, 0], [0, 0, 1]])
        >>> fd2 = skfda.FDataBasis(basis, [1, 0, 0])
        >>> cosine_similarity(fd1, fd2)
        array([ 0.8660254 ,  0.74535599])

        >>> basis = skfda.representation.basis.Monomial(n_basis=3)
        >>>
        >>> fd1 = skfda.FDataBasis(basis, [[0, 1, 0], [0, 0, 1]])
        >>> fd2 = skfda.FDataBasis(basis, [[1, 0, 0], [0, 1, 0]])
        >>> cosine_similarity(fd1, fd2)
        array([ 0.8660254 ,  0.96824584])

    """
    inner_prod = inner_product(arg1, arg2)
    norm1 = np.sqrt(inner_product(arg1, arg1))
    norm2 = np.sqrt(inner_product(arg2, arg2))

    return _clip_cosine(inner_prod / norm1 / norm2)


def cosine_similarity_matrix(
    arg1: Vector,
    arg2: Optional[Vector] = None,
) -> NDArrayFloat:
    """
    Return the cosine similarity matrix between is arguments.

    Args:
        arg1: First sample.
        arg2: Second sample. If it is ``None`` returns the cosine similarity
            between the samples in ``arg1``.

    Returns:
        Cosine similarity matrix between samples.

    See also:
        :func:`cosine_similarity`

    """
    inner_matrix = inner_product_matrix(arg1, arg2)

    if arg2 is None or arg2 is arg1:
        norm1 = np.sqrt(np.diag(inner_matrix))
        norm2 = norm1
    else:
        norm1 = np.sqrt(inner_product(arg1, arg1))
        norm2 = np.sqrt(inner_product(arg2, arg2))

    return _clip_cosine(
        inner_matrix / norm1[:, np.newaxis] / norm2[np.newaxis, :],
    )



from __future__ import annotations

import itertools
import warnings
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted

from ...misc.lstsq import solve_regularized_weighted_lstsq
from ...misc.regularization import L2Regularization, compute_penalty_matrix
from ...representation import FData
from ...representation.basis import Basis
from ._coefficients import CoefficientInfo, coefficient_info_from_covariate

RegularizationType = Union[
    L2Regularization[Any],
    Sequence[Optional[L2Regularization[Any]]],
    None,
]

RegularizationIterableType = Union[
    L2Regularization[Any],
    Iterable[Optional[L2Regularization[Any]]],
    None,
]

AcceptedDataType = Union[
    FData,
    np.ndarray,
]

AcceptedDataCoefsType = Union[
    CoefficientInfo[FData],
    CoefficientInfo[np.ndarray],
]

BasisCoefsType = Sequence[Optional[Basis]]

ArgcheckResultType = Tuple[
    List[AcceptedDataType],
    np.ndarray,
    Optional[np.ndarray],
    List[AcceptedDataCoefsType],
]


class LinearRegression(
    BaseEstimator,  # type: ignore
    RegressorMixin,  # type: ignore
):
    r"""Linear regression with multivariate response.

    This is a regression algorithm equivalent to multivariate linear
    regression, but accepting also functional data expressed in a basis
    expansion.

    The model assumed by this method is:

    .. math::
        y = w_0 + w_1 x_1 + \ldots + w_p x_p + \int w_{p+1}(t) x_{p+1}(t) dt \
        + \ldots + \int w_r(t) x_r(t) dt

    where the covariates can be either multivariate or functional and the
    response is multivariate.

    .. warning::
        For now, only scalar responses are supported.

    Args:
        coef_basis (iterable): Basis of the coefficient functions of the
            functional covariates. If multivariate data is supplied, their
            corresponding entries should be ``None``. If ``None`` is provided
            for a functional covariate, the same basis is assumed. If this
            parameter is ``None`` (the default), it is assumed that ``None``
            is provided for all covariates.
        fit_intercept:  Whether to calculate the intercept for this
            model. If set to False, no intercept will be used in calculations
            (i.e. data is expected to be centered).
        regularization (int, iterable or :class:`Regularization`): If it is
            not a :class:`Regularization` object, linear differential
            operator regularization is assumed. If it
            is an integer, it indicates the order of the
            derivative used in the computing of the penalty matrix. For
            instance 2 means that the differential operator is
            :math:`f''(x)`. If it is an iterable, it consists on coefficients
            representing the differential operator used in the computing of
            the penalty matrix. For instance the tuple (1, 0,
            numpy.sin) means :math:`1 + sin(x)D^{2}`. If not supplied this
            defaults to 2. Only used if penalty_matrix is
            ``None``.

    Attributes:
        coef\_: A list containing the weight coefficient for each
            covariate. For multivariate data, the covariate is a Numpy array.
            For functional data, the covariate is a FDataBasis object.
        intercept\_: Independent term in the linear model. Set to 0.0
            if `fit_intercept = False`.

    Examples:
        >>> from skfda.ml.regression import LinearRegression
        >>> from skfda.representation.basis import (FDataBasis, Monomial,
        ...                                         Constant)

        Multivariate linear regression can be used with functions expressed in
        a basis. Also, a functional basis for the weights can be specified:

        >>> x_basis = Monomial(n_basis=3)
        >>> x_fd = FDataBasis(x_basis, [[0, 0, 1],
        ...                             [0, 1, 0],
        ...                             [0, 1, 1],
        ...                             [1, 0, 1]])
        >>> y = [2, 3, 4, 5]
        >>> linear = LinearRegression()
        >>> _ = linear.fit(x_fd, y)
        >>> linear.coef_[0]
        FDataBasis(
            basis=Monomial(domain_range=((0, 1),), n_basis=3),
            coefficients=[[-15.  96. -90.]],
            ...)
        >>> linear.intercept_
        array([ 1.])
        >>> linear.predict(x_fd)
        array([ 2.,  3.,  4.,  5.])

        Covariates can include also multivariate data:

        >>> x_basis = Monomial(n_basis=2)
        >>> x_fd = FDataBasis(x_basis, [[0, 2],
        ...                             [0, 4],
        ...                             [1, 0],
        ...                             [2, 0],
        ...                             [1, 2],
        ...                             [2, 2]])
        >>> x = [[1, 7], [2, 3], [4, 2], [1, 1], [3, 1], [2, 5]]
        >>> y = [11, 10, 12, 6, 10, 13]
        >>> linear = LinearRegression(
        ...              coef_basis=[None, Constant()])
        >>> _ = linear.fit([x, x_fd], y)
        >>> linear.coef_[0]
        array([ 2.,  1.])
        >>> linear.coef_[1]
        FDataBasis(
        basis=Constant(domain_range=((0, 1),), n_basis=1),
        coefficients=[[ 1.]],
        ...)
        >>> linear.intercept_
        array([ 1.])
        >>> linear.predict([x, x_fd])
        array([ 11.,  10.,  12.,   6.,  10.,  13.])

    """

    def __init__(
        self,
        *,
        coef_basis: Optional[BasisCoefsType] = None,
        fit_intercept: bool = True,
        regularization: RegularizationType = None,
    ) -> None:
        self.coef_basis = coef_basis
        self.fit_intercept = fit_intercept
        self.regularization = regularization

    def fit(  # noqa: D102
        self,
        X: Union[AcceptedDataType, Sequence[AcceptedDataType]],
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ) -> LinearRegression:

        X_new, y, sample_weight, coef_info = self._argcheck_X_y(
            X,
            y,
            sample_weight,
            self.coef_basis,
        )

        regularization: RegularizationIterableType = self.regularization

        if self.fit_intercept:
            new_x = np.ones((len(y), 1))
            X_new = [new_x] + X_new
            coef_info = [coefficient_info_from_covariate(new_x, y)] + coef_info

            if isinstance(regularization, Iterable):
                regularization = itertools.chain([None], regularization)
            elif regularization is not None:
                regularization = (None, regularization)

        inner_products_list = [
            c.regression_matrix(x, y)
            for x, c in zip(X_new, coef_info)
        ]

        # This is C @ J
        inner_products = np.concatenate(inner_products_list, axis=1)

        if sample_weight is not None:
            inner_products = inner_products * np.sqrt(sample_weight)
            y = y * np.sqrt(sample_weight)

        penalty_matrix = compute_penalty_matrix(
            basis_iterable=(c.basis for c in coef_info),
            regularization_parameter=1,
            regularization=regularization,
        )

        if self.fit_intercept and penalty_matrix is not None:
            # Intercept is not penalized
            penalty_matrix[0, 0] = 0

        basiscoefs = solve_regularized_weighted_lstsq(
            coefs=inner_products,
            result=y,
            penalty_matrix=penalty_matrix,
        )

        coef_lengths = np.array([i.shape[1] for i in inner_products_list])
        coef_start = np.cumsum(coef_lengths)
        basiscoef_list = np.split(basiscoefs, coef_start)

        # Express the coefficients in functional form
        coefs = [
            c.convert_from_constant_coefs(bcoefs)
            for c, bcoefs in zip(coef_info, basiscoef_list)
        ]

        if self.fit_intercept:
            self.intercept_ = coefs[0]
            coefs = coefs[1:]
        else:
            self.intercept_ = np.zeros(1)

        self.coef_ = coefs
        self._coef_info = coef_info
        self._target_ndim = y.ndim

        return self

    def predict(  # noqa: D102
        self,
        X: Union[AcceptedDataType, Sequence[AcceptedDataType]],
    ) -> np.ndarray:

        check_is_fitted(self)
        X = self._argcheck_X(X)

        result = np.sum(
            [
                coef_info.inner_product(coef, x)
                for coef, x, coef_info
                in zip(self.coef_, X, self._coef_info)
            ],
            axis=0,
        )

        result += self.intercept_

        if self._target_ndim == 1:
            result = result.ravel()

        return result

    def _argcheck_X(
        self,
        X: Union[AcceptedDataType, Sequence[AcceptedDataType]],
    ) -> Sequence[AcceptedDataType]:
        if isinstance(X, (FData, np.ndarray)):
            X = [X]

        X = [x if isinstance(x, FData) else np.asarray(x) for x in X]

        if all(not isinstance(i, FData) for i in X):
            warnings.warn("All the covariates are scalar.")

        return X

    def _argcheck_X_y(
        self,
        X: Union[AcceptedDataType, Sequence[AcceptedDataType]],
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
        coef_basis: Optional[BasisCoefsType] = None,
    ) -> ArgcheckResultType:
        """Do some checks to types and shapes."""
        # TODO: Add support for Dataframes

        new_X = self._argcheck_X(X)

        if any(isinstance(i, FData) for i in y):
            raise ValueError(
                "Some of the response variables are not scalar",
            )

        y = np.asarray(y)

        if coef_basis is None:
            coef_basis = [None] * len(new_X)

        if len(coef_basis) != len(new_X):
            raise ValueError(
                "Number of regression coefficients does "
                "not match number of independent variables.",
            )

        if any(len(y) != len(x) for x in new_X):
            raise ValueError(
                "The number of samples on independent and "
                "dependent variables should be the same",
            )

        coef_info = [
            coefficient_info_from_covariate(x, y, basis=b)
            for x, b in zip(new_X, coef_basis)
        ]

        if sample_weight is not None:

            if len(sample_weight) != len(y):
                raise ValueError(
                    "The number of sample weights should be "
                    "equal to the number of samples.",
                )

            if np.any(np.array(sample_weight) < 0):
                raise ValueError(
                    "The sample weights should be non negative values",
                )

        return new_X, y, sample_weight, coef_info



from __future__ import annotations

import itertools
import warnings
from typing import Any, Generic, Iterable, Optional, Union

import numpy as np
import scipy.linalg
from sklearn.base import BaseEstimator

from skfda.misc.operators import Identity, gramian_matrix

from ...representation import FData
from ...representation._typing import NDArrayFloat
from ...representation.basis import Basis
from ..operators import Operator
from ..operators._operators import OperatorInput


class L2Regularization(
    BaseEstimator,  # type: ignore
    Generic[OperatorInput],
):
    r"""
    Implements :math:`L_2` (Tikhonov) regularization.

    The penalization term in this type of regularization is the square of the
    :math:`L_2` (Euclidean) norm of a linear operator applied to the function
    or vector

    .. math::
            \lambda \| \Gamma x \|_2^2

    where :math:`\Gamma` is the so called Tikhonov operator
    (matrix for finite vectors) and :math:`\lambda` is a positive real number.

    This linear operator can be an arbitrary Python callable that correspond
    to a linear transformation. However, the
    :doc:`operators </modules/misc/operators>` module
    provides several common linear operators.

    Parameters:
        linear_operator: Linear operator used for regularization. By default
            the second derivative, which is related with the function
            curvature, is penalized.
        regularization_parameter: Scaling parameter (:math:`\lambda`) of the
            penalization.

    Examples:
        Construct a regularization that penalizes the second derivative,
        which is a measure of the curvature of the function.

        >>> from skfda.misc.regularization import L2Regularization
        >>> from skfda.misc.operators import LinearDifferentialOperator
        >>>
        >>> regularization = L2Regularization(
        ...                     LinearDifferentialOperator(2),
        ... )

        By default the regularization penalizes the identity operator:

        >>> regularization = L2Regularization()

        Construct a regularization that penalizes the difference between
        the points :math:`f(1)` and :math:`f(0)` of a function :math:`f`.

        >>> regularization = L2Regularization(lambda x: x(1) - x(0))

        Construct a regularization that penalizes the harmonic acceleration
        operator :math:`Lf = \omega^2 D f + D^3 f`, that, when the
        regularization parameter is large, forces the function to be
        :math:`f(t) = c_1 + c_2 \sin \omega t + c_3 \cos \omega t`, where
        :math:`\omega` is the angular frequency. This is useful for some
        periodic functions.

        >>> import numpy as np
        >>>
        >>> period = 1
        >>> w = 2 * np.pi / period
        >>> regularization = L2Regularization(
        ...                     LinearDifferentialOperator([0, w**2, 0, 1]),
        ... )

    """

    def __init__(
        self,
        linear_operator: Optional[Operator[OperatorInput, Any]] = None,
        *,
        regularization_parameter: float = 1,
    ) -> None:
        self.linear_operator = linear_operator
        self.regularization_parameter = regularization_parameter

    def penalty_matrix(
        self,
        basis: OperatorInput,
    ) -> NDArrayFloat:
        """Return a penalty matrix for ordinary least squares."""
        linear_operator = (
            Identity()
            if self.linear_operator is None
            else self.linear_operator
        )

        return self.regularization_parameter * gramian_matrix(
            linear_operator,
            basis,
        )


class TikhonovRegularization(
    L2Regularization[OperatorInput],
):

    def __init__(
        self,
        linear_operator: Optional[Operator[OperatorInput, Any]] = None,
        *,
        regularization_parameter: float = 1,
    ) -> None:

        warnings.warn(
            "Class TikhonovRegularization is deprecated. Use class "
            "L2Regularization instead.",
            DeprecationWarning,
        )

        return super().__init__(
            linear_operator=linear_operator,
            regularization_parameter=regularization_parameter,
        )


BasisTypes = Union[np.ndarray, FData, Basis]
Regularization = L2Regularization[Any]
RegularizationLike = Union[
    None,
    Regularization,
    Iterable[Optional[Regularization]],
]


def compute_penalty_matrix(
    basis_iterable: Iterable[BasisTypes],
    regularization_parameter: Union[float, Iterable[float]],
    regularization: RegularizationLike,
) -> Optional[NDArrayFloat]:
    """
    Compute the regularization matrix for a linear differential operator.

    X can be a list of mixed data.

    """
    # If there is no regularization, return 0 and rely on broadcasting
    if regularization_parameter == 0 or regularization is None:
        return None

    # Compute penalty matrix if not provided
    if not isinstance(regularization, Iterable):
        regularization = (regularization,)

    if not isinstance(regularization_parameter, Iterable):
        regularization_parameter = itertools.repeat(
            regularization_parameter,
        )

    penalty_blocks = [
        np.zeros((len(b), len(b))) if r is None else
        a * r.penalty_matrix(b)
        for b, r, a in zip(
            basis_iterable,
            regularization,
            regularization_parameter,
        )]

    return scipy.linalg.block_diag(*penalty_blocks)



"""Common types."""
from typing import Any, Optional, Sequence, Tuple, TypeVar, Union

from typing_extensions import Protocol

import numpy as np

try:
    from numpy.typing import ArrayLike
except ImportError:
    ArrayLike = np.ndarray  # type:ignore

try:
    from numpy.typing import NDArray
    NDArrayAny = NDArray[Any]
    NDArrayInt = NDArray[np.int_]
    NDArrayFloat = NDArray[np.float_]
    NDArrayBool = NDArray[np.bool_]
except ImportError:
    NDArray = np.ndarray  # type:ignore
    NDArrayAny = np.ndarray  # type:ignore
    NDArrayInt = np.ndarray  # type:ignore
    NDArrayFloat = np.ndarray  # type:ignore
    NDArrayBool = np.ndarray  # type:ignore

VectorType = TypeVar("VectorType")

DomainRange = Tuple[Tuple[float, float], ...]
DomainRangeLike = Union[
    DomainRange,
    Sequence[float],
    Sequence[Sequence[float]],
]

LabelTuple = Tuple[Optional[str], ...]
LabelTupleLike = Sequence[Optional[str]]

GridPoints = Tuple[np.ndarray, ...]
GridPointsLike = Union[ArrayLike, Sequence[ArrayLike]]


class Vector(Protocol):
    """
    Protocol representing a generic vector.

    It should accept numpy arrays and FData, among other things.
    """

    def __add__(
        self: VectorType,
        __other: VectorType,  # noqa: WPS112
    ) -> VectorType:
        pass

    def __sub__(
        self: VectorType,
        __other: VectorType,  # noqa: WPS112
    ) -> VectorType:
        pass

    def __mul__(
        self: VectorType,
        __other: float,  # noqa: WPS112
    ) -> VectorType:
        pass



from __future__ import annotations

from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator, OutlierMixin

from ...representation import FDataGrid
from ..depth import Depth, ModifiedBandDepth
from . import _envelopes


class BoxplotOutlierDetector(
    BaseEstimator,  # type: ignore
    OutlierMixin,  # type: ignore
):
    r"""Outlier detector using the interquartile range.

    Detects as outliers functions that have one or more points outside
    ``factor`` times the interquartile range plus or minus the central
    envelope, given a functional depth measure. This corresponds to the
    points selected as outliers by the functional boxplot.

    Parameters:
        depth_method (Callable): The functional depth measure used.
        factor (float): The number of times the IQR is multiplied.

    Example:
        Function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> import skfda
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> out_detector = BoxplotOutlierDetector()
        >>> out_detector.fit_predict(fd)
        array([-1, 1, 1, -1])

    """

    def __init__(
        self,
        *,
        depth_method: Optional[Depth[FDataGrid]] = None,
        factor: float = 1.5,
    ) -> None:
        self.depth_method = depth_method
        self.factor = factor

    def fit(self, X: FDataGrid, y: None = None) -> BoxplotOutlierDetector:

        depth_method = (
            self.depth_method
            if self.depth_method is not None
            else ModifiedBandDepth()
        )
        depth = depth_method(X)
        indices_descending_depth = (-depth).argsort(axis=0)

        # Central region and envelope must be computed for outlier detection
        central_region = _envelopes.compute_region(
            X, indices_descending_depth, 0.5)
        self._central_envelope = _envelopes.compute_envelope(central_region)

        # Non-outlying envelope
        self.non_outlying_threshold_ = _envelopes.non_outlying_threshold(
            self._central_envelope, self.factor)

        return self

    def predict(self, X: FDataGrid) -> np.ndarray:
        outliers = _envelopes.predict_outliers(
            X,
            self.non_outlying_threshold_,
        )

        # Predict as scikit-learn outlier detectors
        predicted = ~outliers + outliers * -1

        return predicted



"""Abstract base class for basis."""

from __future__ import annotations

import copy
import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Tuple, TypeVar, Union

import numpy as np
from matplotlib.figure import Figure

from ..._utils import _reshape_eval_points, _same_domain, _to_domain_range
from .._typing import ArrayLike, DomainRange, DomainRangeLike

if TYPE_CHECKING:
    from . import FDataBasis

T = TypeVar("T", bound='Basis')


class Basis(ABC):
    """Defines the structure of a basis of functions.

    Parameters:
        domain_range: The :term:`domain range` over which the basis can be
            evaluated.
        n_basis: number of functions in the basis.

    """

    def __init__(
        self,
        *,
        domain_range: Optional[DomainRangeLike] = None,
        n_basis: int = 1,
    ) -> None:
        """Basis constructor."""
        if domain_range is not None:

            domain_range = _to_domain_range(domain_range)

        if n_basis < 1:
            raise ValueError(
                "The number of basis has to be strictly positive.",
            )

        self._domain_range = domain_range
        self._n_basis = n_basis

        super().__init__()

    def __call__(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
    ) -> np.ndarray:
        """Evaluate Basis objects.

        Evaluates the basis functions at a list of given values.

        Args:
            eval_points: List of points where the basis is
                evaluated.
            derivative: order of the derivative.

                .. deprecated:: 0.4
                    Use `derivative` method instead.

        Returns:
            Matrix whose rows are the values of the each
            basis function or its derivatives at the values specified in
            eval_points.

        """
        return self.evaluate(eval_points, derivative=derivative)

    @property
    def dim_domain(self) -> int:
        if self._domain_range is None:
            return 1
        return len(self._domain_range)

    @property
    def dim_codomain(self) -> int:
        return 1

    @property
    def domain_range(self) -> DomainRange:
        if self._domain_range is None:
            return ((0, 1),) * self.dim_domain

        return self._domain_range

    @property
    def n_basis(self) -> int:
        return self._n_basis

    def is_domain_range_fixed(self) -> bool:
        """
        Return wether the :term:`domain range` has been set explicitly.

        This is useful when using a basis for converting a dataset, since
        if this is not explicitly assigned it can be changed to the domain of
        the data.

        Returns:
            `True` if the domain range has been fixed. `False` otherwise.

        """
        return self._domain_range is not None

    @abstractmethod
    def _evaluate(
        self,
        eval_points: np.ndarray,
    ) -> np.ndarray:
        """
        Evaluate Basis object.

        Subclasses must override this to provide basis evaluation.

        """
        pass

    def evaluate(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
    ) -> np.ndarray:
        """Evaluate Basis objects and its derivatives.

        Evaluates the basis functions at a list of given values.

        Args:
            eval_points: List of points where the basis is
                evaluated.
            derivative: order of the derivative.

                .. deprecated:: 0.4
                    Use `derivative` method instead.

        Returns:
            Matrix whose rows are the values of the each
            basis function or its derivatives at the values specified in
            eval_points.

        """
        if derivative < 0:
            raise ValueError("derivative only takes non-negative values.")
        elif derivative != 0:
            warnings.warn(
                "Parameter derivative is deprecated. Use the "
                "derivative method instead.",
                DeprecationWarning,
            )
            return self.derivative(order=derivative)(eval_points)

        eval_points = _reshape_eval_points(
            eval_points,
            aligned=True,
            n_samples=self.n_basis,
            dim_domain=self.dim_domain,
        )

        return self._evaluate(eval_points).reshape(
            (self.n_basis, len(eval_points), self.dim_codomain),
        )

    def __len__(self) -> int:
        return self.n_basis

    def derivative(self, *, order: int = 1) -> FDataBasis:
        """Construct a FDataBasis object containing the derivative.

        Args:
            order: Order of the derivative. Defaults to 1.

        Returns:
            Derivative object.

        """
        return self.to_basis().derivative(order=order)

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:
        """
        Return basis and coefficients of the derivative.

        Args:
            coefs: Coefficients of a vector expressed in this basis.
            order: Order of the derivative.

        Returns:
            Tuple with the basis of the derivative and its coefficients.

        Subclasses can override this to provide derivative construction.

        """
        raise NotImplementedError(
            f"{type(self)} basis does not support the construction of a "
            "basis of the derivatives.",
        )

    def derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:
        """
        Return basis and coefficients of the derivative.

        Args:
            coefs: Coefficients of a vector expressed in this basis.
            order: Order of the derivative.

        Returns:
            Tuple with the basis of the derivative and its coefficients.

        """
        return self._derivative_basis_and_coefs(coefs, order)

    def plot(self, *args: Any, **kwargs: Any) -> Figure:
        """Plot the basis object or its derivatives.

        Args:
            args: arguments to be passed to the
                fdata.plot function.
            kwargs: keyword arguments to be passed to the
                fdata.plot function.

        Returns:
            Figure object in which the graphs are plotted.

        """
        self.to_basis().plot(*args, **kwargs)

    def _coordinate_nonfull(
        self,
        coefs: np.ndarray,
        key: Union[int, slice],
    ) -> Tuple[Basis, np.ndarray]:
        """
        Return a basis and coefficients for the indexed coordinate functions.

        Subclasses can override this to provide coordinate indexing.

        """
        raise NotImplementedError("Coordinate indexing not implemented")

    def coordinate_basis_and_coefs(
        self,
        coefs: np.ndarray,
        key: Union[int, slice],
    ) -> Tuple[Basis, np.ndarray]:
        """Return a fdatabasis for the coordinate functions indexed by key."""
        # Raises error if not in range and normalize key
        r_key = range(self.dim_codomain)[key]

        if isinstance(r_key, range) and len(r_key) == 0:
            raise IndexError("Empty number of coordinates selected")

        # Full fdatabasis case
        if (
            (self.dim_codomain == 1 and r_key == 0)
            or (isinstance(r_key, range) and len(r_key) == self.dim_codomain)
        ):
            return self, np.copy(coefs)

        return self._coordinate_nonfull(
            coefs=coefs,
            key=key,
        )

    def rescale(self: T, domain_range: Optional[DomainRangeLike] = None) -> T:
        """
        Return a copy of the basis with a new :term:`domain` range.

        Args:
            domain_range: Definition of the interval
                where the basis defines a space. Defaults uses the same as
                the original basis.

        Returns:
            Rescaled copy.

        """
        return self.copy(domain_range=domain_range)

    def copy(self: T, domain_range: Optional[DomainRangeLike] = None) -> T:
        """Basis copy."""
        new_copy = copy.deepcopy(self)

        if domain_range is not None:
            domain_range = _to_domain_range(domain_range)

            new_copy._domain_range = domain_range  # noqa: WPS437

        return new_copy

    def to_basis(self) -> FDataBasis:
        """
        Convert the Basis to FDatabasis.

        Returns:
            FDataBasis with this basis as its basis, and all basis functions
            as observations.

        """
        from . import FDataBasis
        return FDataBasis(self.copy(), np.identity(self.n_basis))

    def _to_R(self) -> str:  # noqa: N802
        raise NotImplementedError

    def inner_product_matrix(
        self,
        other: Optional[Basis] = None,
    ) -> np.ndarray:
        r"""
        Return the Inner Product Matrix of a pair of basis.

        The Inner Product Matrix is defined as

        .. math::
            I_{ij} = \langle\phi_i, \theta_j\rangle

        where :math:`\phi_i` is the ith element of the basis and
        :math:`\theta_j` is the jth element of the second basis.
        This matrix helps on the calculation of the inner product
        between objects on two basis and for the change of basis.

        Args:
            other: Basis to compute the inner product
                matrix. If not basis is given, it computes the matrix with
                itself returning the Gram Matrix

        Returns:
            Inner Product Matrix of two basis

        """
        from ...misc import inner_product_matrix

        if other is None or self == other:
            return self.gram_matrix()

        return inner_product_matrix(self, other)

    def _gram_matrix_numerical(self) -> np.ndarray:
        """Compute the Gram matrix numerically."""
        from ...misc import inner_product_matrix

        return inner_product_matrix(self, force_numerical=True)

    def _gram_matrix(self) -> np.ndarray:
        """
        Compute the Gram matrix.

        Subclasses may override this method for improving computation
        of the Gram matrix.

        """
        return self._gram_matrix_numerical()

    def gram_matrix(self) -> np.ndarray:
        r"""
        Return the Gram Matrix of a basis.

        The Gram Matrix is defined as

        .. math::
            G_{ij} = \langle\phi_i, \phi_j\rangle

        where :math:`\phi_i` is the ith element of the basis. This is a
        symmetric matrix and positive-semidefinite.

        Returns:
            Gram Matrix of the basis.

        """
        gram = getattr(self, "_gram_matrix_cached", None)

        if gram is None:
            gram = self._gram_matrix()
            self._gram_matrix_cached = gram

        return gram

    def _mul_constant(
        self: T,
        coefs: np.ndarray,
        other: float,
    ) -> Tuple[T, np.ndarray]:
        coefs = coefs.copy()
        other = np.atleast_2d(other).reshape(-1, 1)
        coefs *= other

        return self.copy(), coefs

    def __repr__(self) -> str:
        """Representation of a Basis object."""
        return (
            f"{self.__class__.__name__}("
            f"domain_range={self.domain_range}, "
            f"n_basis={self.n_basis})"
        )

    def __eq__(self, other: Any) -> bool:
        """Test equality of Basis."""
        return (
            isinstance(other, type(self))
            and _same_domain(self, other)
            and self.n_basis == other.n_basis
        )

    def __hash__(self) -> int:
        """Hash a Basis."""
        return hash((self.domain_range, self.n_basis))



import unittest

import pandas as pd

import skfda


class TestPandas(unittest.TestCase):

    def setUp(self) -> None:
        self.fd = skfda.FDataGrid(
            [[1, 2, 3, 4, 5, 6, 7], [2, 3, 4, 5, 6, 7, 9]])
        self.fd_basis = self.fd.to_basis(skfda.representation.basis.BSpline(
            n_basis=5))

    def test_fdatagrid_series(self) -> None:
        series = pd.Series(self.fd)
        self.assertIsInstance(
            series.dtype, skfda.representation.grid.FDataGridDType)
        self.assertEqual(len(series), self.fd.n_samples)
        self.assertTrue(series[0].equals(self.fd[0]))

    def test_fdatabasis_series(self) -> None:
        series = pd.Series(self.fd_basis)
        self.assertIsInstance(
            series.dtype, skfda.representation.basis.FDataBasisDType)
        self.assertEqual(len(series), self.fd_basis.n_samples)
        self.assertTrue(series[0].equals(self.fd_basis[0]))

    def test_fdatagrid_dataframe(self) -> None:
        df = pd.DataFrame({"function": self.fd})
        self.assertIsInstance(
            df["function"].dtype, skfda.representation.grid.FDataGridDType)
        self.assertEqual(len(df["function"]), self.fd.n_samples)
        self.assertTrue(df["function"][0].equals(self.fd[0]))

    def test_fdatabasis_dataframe(self) -> None:
        df = pd.DataFrame({"function": self.fd_basis})
        self.assertIsInstance(
            df["function"].dtype, skfda.representation.basis.FDataBasisDType)
        self.assertEqual(len(df["function"]), self.fd_basis.n_samples)
        self.assertTrue(df["function"][0].equals(self.fd_basis[0]))

    def test_take(self) -> None:
        self.assertTrue(self.fd.take(0).equals(self.fd[0]))
        self.assertTrue(self.fd.take(0, axis=0).equals(self.fd[0]))



from typing import Optional, Tuple, TypeVar

import numpy as np

from .._typing import DomainRangeLike
from ._basis import Basis

T = TypeVar("T", bound='Constant')


class Constant(Basis):
    """Constant basis.

    Basis for constant functions

    Parameters:
        domain_range: The :term:`domain range` over which the basis can be
            evaluated.

    Examples:
        Defines a contant base over the interval :math:`[0, 5]` consisting
        on the constant function 1 on :math:`[0, 5]`.

        >>> bs_cons = Constant((0,5))

    """

    def __init__(self, domain_range: Optional[DomainRangeLike] = None) -> None:
        """Constant basis constructor."""
        super().__init__(domain_range=domain_range, n_basis=1)

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:
        return np.ones((1, len(eval_points)))

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:
        return (
            (self.copy(), coefs.copy()) if order == 0
            else (self.copy(), np.zeros(coefs.shape))
        )

    def _gram_matrix(self) -> np.ndarray:
        return np.array(
            [[self.domain_range[0][1] - self.domain_range[0][0]]],
        )

    def _to_R(self) -> str:  # noqa: N802
        drange = self.domain_range[0]
        drange_str = f"c({str(drange[0])}, {str(drange[1])})"
        return f"create.constant.basis(rangeval = {drange_str})"



"""Defines methods for the validation of the smoothing."""
import numpy as np
import sklearn
from sklearn.model_selection import GridSearchCV

__author__ = "Miguel Carbajo Berrocal"
__email__ = "miguel.carbajo@estudiante.uam.es"


def _get_input_estimation_and_matrix(estimator, X):
    """Returns the smoothed data evaluated at the input points & the matrix"""
    if estimator.output_points is not None:
        estimator = sklearn.base.clone(estimator)
        estimator.output_points = None
        estimator.fit(X)
    y_est = estimator.transform(X)

    hat_matrix = estimator.hat_matrix_

    return y_est, hat_matrix


class LinearSmootherLeaveOneOutScorer():
    r"""Leave-one-out cross validation scoring method for linear smoothers.

    It calculates the cross validation score for every sample in a FDataGrid
    object given a linear smoother with a smoothing matrix :math:`\hat{H}^\nu`
    calculated with a parameter :math:`\nu`:

    .. math::
        CV(\nu)=\frac{1}{n} \sum_i \left(y_i - \hat{y}_i^{\nu(
        -i)}\right)^2

    Where :math:`\hat{y}_i^{\nu(-i)}` is the adjusted :math:`y_i` when the
    the pair of values :math:`(x_i,y_i)` are excluded in the smoothing. This
    would require to recalculate the smoothing matrix n times. Fortunately
    the above formula can be expressed in a way where the smoothing matrix
    does not need to be calculated again.

    .. math::
        CV(\nu)=\frac{1}{n} \sum_i \left(\frac{y_i - \hat{y}_i^\nu}{1 -
        \hat{H}_{ii}^\nu}\right)^2

    Args:
        estimator (Estimator): Linear smoothing estimator.
        X (FDataGrid): Functional data to smooth.
        y (FDataGrid): Functional data target. Should be the same as X.

    Returns:
        float: Cross validation score, with negative sign, as it is a
        penalization.

    """

    def __call__(self, estimator, X, y):

        y_est, hat_matrix = _get_input_estimation_and_matrix(estimator, X)

        return -np.mean(((y.data_matrix[..., 0] - y_est.data_matrix[..., 0])
                         / (1 - hat_matrix.diagonal())) ** 2)


class LinearSmootherGeneralizedCVScorer():
    r"""Generalized cross validation scoring method for linear smoothers.

    It calculates the general cross validation score for every sample in a
    FDataGrid object given a smoothing matrix :math:`\hat{H}^\nu`
    calculated with a parameter :math:`\nu`:

    .. math::
        GCV(\nu)=\Xi(\nu,n)\frac{1}{n} \sum_i \left(y_i - \hat{
        y}_i^\nu\right)^2

    Where :math:`\hat{y}_i^{\nu}` is the adjusted :math:`y_i` and
    :math:`\Xi` is a penalization function. By default the penalization
    function is:

    .. math::
        \Xi(\nu,n) = \left( 1 - \frac{tr(\hat{H}^\nu)}{n} \right)^{-2}

    but others such as the Akaike's information criterion can be considered.

    Args:
        estimator (Estimator): Linear smoothing estimator.
        X (FDataGrid): Functional data to smooth.
        y (FDataGrid): Functional data target. Should be the same as X.

    Returns:
        float: Cross validation score, with negative sign, as it is a
        penalization.

    """

    def __init__(self, penalization_function=None):
        self.penalization_function = penalization_function

    def __call__(self, estimator, X, y):
        y_est, hat_matrix = _get_input_estimation_and_matrix(estimator, X)

        if self.penalization_function is None:
            def penalization_function(hat_matrix):
                return (1 - hat_matrix.diagonal().mean()) ** -2
        else:
            penalization_function = self.penalization_function

        return -(np.mean(((y.data_matrix[..., 0] - y_est.data_matrix[..., 0])
                          / (1 - hat_matrix.diagonal())) ** 2)
                 * penalization_function(hat_matrix))


class SmoothingParameterSearch(GridSearchCV):
    """Chooses the best smoothing parameter and performs smoothing.

    Performs the smoothing of a FDataGrid object choosing the best
    parameter of a given list using a cross validation scoring method.

    Note:
        This is similar to fitting a scikit-learn GridSearchCV over the
        data, using the cv_method as a scorer.

    Args:
        estimator (smoother estimator): scikit-learn compatible smoother.
        param_values (iterable): iterable containing the values to test
            for *smoothing_parameter*.
        scoring (scoring method): scoring method used to measure the
            performance of the smoothing. If ``None`` (the default) the
            ``score`` method of the estimator is used.
        n_jobs (int or None, optional (default=None)):
            Number of jobs to run in parallel.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context. ``-1`` means using all processors. See
            :term:`scikit-learn Glossary <sklearn:n_jobs>` for more details.

        pre_dispatch (int, or string, optional):
            Controls the number of jobs that get dispatched during parallel
            execution. Reducing this number can be useful to avoid an
            explosion of memory consumption when more jobs get dispatched
            than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A string, giving an expression as a function of n_jobs,
              as in '2*n_jobs'
        verbose (integer):
            Controls the verbosity: the higher, the more messages.

        error_score ('raise' or numeric):
            Value to assign to the score if an error occurs in estimator
            fitting. If set to 'raise', the error is raised. If a numeric
            value is given, FitFailedWarning is raised. This parameter does
            not affect the refit step, which will always raise the error.
            Default is np.nan.

    Examples:
        Creates a FDataGrid object of the function :math:`y=x^2` and peforms
        smoothing by means of the k-nearest neighbours method.

        >>> import skfda
        >>> from skfda.preprocessing.smoothing import kernel_smoothers
        >>> x = np.linspace(-2, 2, 5)
        >>> fd = skfda.FDataGrid(x ** 2, x)
        >>> grid = SmoothingParameterSearch(
        ...            kernel_smoothers.KNeighborsSmoother(), [2,3])
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([-11.67, -12.37])
        >>> round(grid.best_score_, 2)
        -11.67
        >>> grid.best_params_['smoothing_parameter']
        2
        >>> grid.best_estimator_.hat_matrix().round(2)
        array([[ 0.5 , 0.5 , 0.  , 0.  , 0.  ],
               [ 0.33, 0.33, 0.33, 0.  , 0.  ],
               [ 0.  , 0.33, 0.33, 0.33, 0.  ],
               [ 0.  , 0.  , 0.33, 0.33, 0.33],
               [ 0.  , 0.  , 0.  , 0.5 , 0.5 ]])
        >>> grid.transform(fd).round(2)
        FDataGrid(
            array([[[ 2.5 ],
                    [ 1.67],
                    [ 0.67],
                    [ 1.67],
                    [ 2.5 ]]]),
            grid_points=(array([-2., -1.,  0.,  1.,  2.]),),
            domain_range=((-2.0, 2.0),),
            ...)

        Other validation methods can be used such as cross-validation or
        general cross validation using other penalization functions.

        >>> grid = SmoothingParameterSearch(
        ...         kernel_smoothers.KNeighborsSmoother(), [2,3],
        ...         scoring=LinearSmootherLeaveOneOutScorer())
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([-4.2, -5.5])
        >>> grid = SmoothingParameterSearch(
        ...         kernel_smoothers.KNeighborsSmoother(), [2,3],
        ...         scoring=LinearSmootherGeneralizedCVScorer(
        ...                         akaike_information_criterion))
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([ -9.35, -10.71])
        >>> grid = SmoothingParameterSearch(
        ...         kernel_smoothers.KNeighborsSmoother(), [2,3],
        ...         scoring=LinearSmootherGeneralizedCVScorer(
        ...                         finite_prediction_error))
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([ -9.8, -11. ])
        >>> grid = SmoothingParameterSearch(
        ...         kernel_smoothers.KNeighborsSmoother(), [2,3],
        ...         scoring=LinearSmootherGeneralizedCVScorer(shibata))
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([-7.56, -9.17])
        >>> grid = SmoothingParameterSearch(
        ...         kernel_smoothers.KNeighborsSmoother(), [2,3],
        ...         scoring=LinearSmootherGeneralizedCVScorer(rice))
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([-21. , -16.5])

        Different output points can also be used. In that case the value used
        as a target is still the smoothed value at the input points:

        >>> output_points = np.linspace(-2, 2, 9)
        >>> grid = SmoothingParameterSearch(
        ...            kernel_smoothers.KNeighborsSmoother(
        ...                output_points=output_points
        ...            ), [2,3])
        >>> _ = grid.fit(fd)
        >>> np.array(grid.cv_results_['mean_test_score']).round(2)
        array([-11.67, -12.37])
        >>> grid.transform(fd).data_matrix.round(2)
        array([[[ 2.5 ],
                [ 2.5 ],
                [ 1.67],
                [ 0.5 ],
                [ 0.67],
                [ 0.5 ],
                [ 1.67],
                [ 2.5 ],
                [ 2.5 ]]])
    """

    def __init__(self, estimator, param_values, *, scoring=None, n_jobs=None,
                 verbose=0, pre_dispatch='2*n_jobs',
                 error_score=np.nan):
        super().__init__(estimator=estimator, scoring=scoring,
                         param_grid={'smoothing_parameter': param_values},
                         n_jobs=n_jobs,
                         refit=True, cv=[(slice(None), slice(None))],
                         verbose=verbose, pre_dispatch=pre_dispatch,
                         error_score=error_score, return_train_score=False)
        self.param_values = param_values

    def fit(self, X, y=None, groups=None, **fit_params):
        if y is None:
            y = X

        return super().fit(X, y=y, groups=groups, **fit_params)


def akaike_information_criterion(hat_matrix):
    r"""Akaike's information criterion for cross validation.

    .. math::
        \Xi(\nu,n) = \exp\left(2 * \frac{tr(\hat{H}^\nu)}{n}\right)

    Args:
        hat_matrix (numpy.darray): Smoothing matrix whose penalization
            score is desired.

    Returns:
         float: penalization given by the Akaike's information criterion.

    """
    return np.exp(2 * hat_matrix.diagonal().mean())


def finite_prediction_error(hat_matrix):
    r"""Finite prediction error for cross validation.

    .. math::
        \Xi(\nu,n) = \frac{1 + \frac{tr(\hat{H}^\nu)}{n}}{1 -
        \frac{tr(\hat{H}^\nu)}{n}}

    Args:
        hat_matrix (numpy.darray): Smoothing matrix whose penalization
            score is desired.

    Returns:
         float: penalization given by the finite prediction error.

    """
    return ((1 + hat_matrix.diagonal().mean())
            / (1 - hat_matrix.diagonal().mean()))


def shibata(hat_matrix):
    r"""Shibata's model selector for cross validation.

    .. math::
        \Xi(\nu,n) = 1 + 2 * \frac{tr(\hat{H}^\nu)}{n}

    Args:
        hat_matrix (numpy.darray): Smoothing matrix whose penalization
            score is desired.

    Returns:
         float: penalization given by the Shibata's model selector.

    """
    return 1 + 2 * hat_matrix.diagonal().mean()


def rice(hat_matrix):
    r"""Rice's bandwidth selector for cross validation.

    .. math::
        \Xi(\nu,n) = \left(1 - 2 * \frac{tr(\hat{H}^\nu)}{n}\right)^{-1}

    Args:
        hat_matrix (numpy.darray): Smoothing matrix whose penalization
            score is desired.

    Returns:
         float: penalization given by the Rice's bandwidth selector.

    """
    return (1 - 2 * hat_matrix.diagonal().mean()) ** -1




from skfda.representation.basis import (
    FDataBasis, Monomial, BSpline, Fourier, Constant, VectorValued, Tensor)
import unittest

import numpy as np


class TestBasisEvaluationFourier(unittest.TestCase):

    def test_evaluation_simple_fourier(self):
        """Test the evaluation of FDataBasis"""
        fourier = Fourier(domain_range=(0, 2), n_basis=5)

        coefficients = np.array([[1,  2,  3,  4,  5],
                                 [6,  7,  8,  9, 10]])

        f = FDataBasis(fourier, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array([[8.71,  9.66,  1.84, -4.71, -2.80, 2.71,
                         2.45, -3.82,  -6.66, -0.30,  8.71],
                        [22.24, 26.48, 10.57, -4.95, -3.58, 6.24,
                         5.31, -7.69, -13.32,  1.13, 22.24]])[..., np.newaxis]

        np.testing.assert_array_almost_equal(f(t).round(2), res)
        np.testing.assert_array_almost_equal(f.evaluate(t).round(2), res)

    def test_evaluation_point_fourier(self):
        """Test the evaluation of a single point FDataBasis"""
        fourier = Fourier(domain_range=(0, 1), n_basis=3)

        coefficients = np.array([[0.00078238, 0.48857741, 0.63971985],
                                 [0.01778079, 0.73440271, 0.20148638]])

        f = FDataBasis(fourier, coefficients)

        # Test different ways of call f with a point
        res = np.array([-0.903918107989282, -0.267163981229459]
                       ).reshape((2, 1, 1)).round(4)

        np.testing.assert_array_almost_equal(f([0.5]).round(4), res)
        np.testing.assert_array_almost_equal(f((0.5,)).round(4), res)
        np.testing.assert_array_almost_equal(f(0.5).round(4), res)
        np.testing.assert_array_almost_equal(f(np.array([0.5])).round(4), res)

        # Problematic case, should be accepted or no?
        #np.testing.assert_array_almost_equal(f(np.array(0.5)).round(4), res)

    def test_evaluation_derivative_fourier(self):
        """Test the evaluation of the derivative of a FDataBasis"""
        fourier = Fourier(domain_range=(0, 1), n_basis=3)

        coefficients = np.array([[0.00078238, 0.48857741, 0.63971985],
                                 [0.01778079, 0.73440271, 0.20148638]])

        f = FDataBasis(fourier, coefficients)

        t = np.linspace(0, 1, 4)

        res = np.array([4.34138447771721, -7.09352774867064, 2.75214327095343,
                        4.34138447771721, 6.52573053999253,
                        -4.81336320468984, -1.7123673353027, 6.52573053999253]
                       ).reshape((2, 4, 1)).round(3)

        f_deriv = f.derivative()
        np.testing.assert_array_almost_equal(
            f_deriv(t).round(3), res
        )

    def test_evaluation_grid_fourier(self):
        """Test the evaluation of FDataBasis with the grid option set to
            true. Nothing should be change due to the domain dimension is 1,
            but can accept the """
        fourier = Fourier(domain_range=(0, 1), n_basis=3)

        coefficients = np.array([[0.00078238, 0.48857741, 0.63971985],
                                 [0.01778079, 0.73440271, 0.20148638]])

        f = FDataBasis(fourier, coefficients)
        t = np.linspace(0, 1, 4)

        res_test = f(t)

        # Different ways to pass the axes
        np.testing.assert_array_almost_equal(f(t, grid=True), res_test)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res_test)
        np.testing.assert_array_almost_equal(f([t], grid=True), res_test)
        np.testing.assert_array_almost_equal(f(np.atleast_2d(t), grid=True),
                                             res_test)

        # Number of axis different than the domain dimension (1)
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_composed_fourier(self):
        """Test the evaluation of FDataBasis the a matrix of times instead of
        a list of times """
        fourier = Fourier(domain_range=(0, 1), n_basis=3)

        coefficients = np.array([[0.00078238, 0.48857741, 0.63971985],
                                 [0.01778079, 0.73440271, 0.20148638]])

        f = FDataBasis(fourier, coefficients)
        t = np.linspace(0, 1, 4)

        # Test same result than evaluation standart
        np.testing.assert_array_almost_equal(f([1]),
                                             f([[1], [1]],
                                               aligned=False))
        np.testing.assert_array_almost_equal(f(t), f(np.vstack((t, t)),
                                                     aligned=False))

        # Different evaluation times
        t_multiple = [[0, 0.5], [0.2, 0.7]]
        np.testing.assert_array_almost_equal(f(t_multiple[0])[0],
                                             f(t_multiple,
                                               aligned=False)[0])
        np.testing.assert_array_almost_equal(f(t_multiple[1])[1],
                                             f(t_multiple,
                                               aligned=False)[1])

    def test_domain_in_list_fourier(self):
        """Test the evaluation of FDataBasis"""
        for fourier in (Fourier(domain_range=[(0, 1)], n_basis=3),
                        Fourier(domain_range=((0, 1),), n_basis=3),
                        Fourier(domain_range=np.array((0, 1)), n_basis=3),
                        Fourier(domain_range=np.array([(0, 1)]), n_basis=3)):

            coefficients = np.array([[0.00078238, 0.48857741, 0.63971985],
                                     [0.01778079, 0.73440271, 0.20148638]])

            f = FDataBasis(fourier, coefficients)

            t = np.linspace(0, 1, 4)

            res = np.array([0.905, 0.147, -1.05, 0.905, 0.303,
                            0.775, -1.024, 0.303]).reshape((2, 4, 1))

            np.testing.assert_array_almost_equal(f(t).round(3), res)
            np.testing.assert_array_almost_equal(f.evaluate(t).round(3), res)


class TestBasisEvaluationBSpline(unittest.TestCase):

    def test_evaluation_simple_bspline(self):
        """Test the evaluation of FDataBasis"""
        bspline = BSpline(domain_range=(0, 2), n_basis=5)

        coefficients = np.array([[1,  2,  3,  4,  5],
                                 [6,  7,  8,  9, 10]])

        f = FDataBasis(bspline, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array([[1, 1.54, 1.99, 2.37, 2.7, 3,
                         3.3, 3.63, 4.01, 4.46, 5],
                        [6, 6.54, 6.99, 7.37, 7.7, 8,
                         8.3, 8.63, 9.01, 9.46, 10]])[..., np.newaxis]

        np.testing.assert_array_almost_equal(f(t).round(2), res)
        np.testing.assert_array_almost_equal(f.evaluate(t).round(2), res)

    def test_evaluation_point_bspline(self):
        """Test the evaluation of a single point FDataBasis"""
        bspline = BSpline(domain_range=(0, 1), n_basis=5, order=3)

        coefficients = [[0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
                        [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12]]

        f = FDataBasis(bspline, coefficients)

        # Test different ways of call f with a point
        res = np.array([[0.5696], [0.3104]])[..., np.newaxis]

        np.testing.assert_array_almost_equal(f([0.5]).round(4), res)
        np.testing.assert_array_almost_equal(f((0.5,)).round(4), res)
        np.testing.assert_array_almost_equal(f(0.5).round(4), res)
        np.testing.assert_array_almost_equal(f(np.array([0.5])).round(4), res)

        # Problematic case, should be accepted or no?
        #np.testing.assert_array_almost_equal(f(np.array(0.5)).round(4), res)

    def test_evaluation_derivative_bspline(self):
        """Test the evaluation of the derivative of a FDataBasis"""
        bspline = BSpline(domain_range=(0, 1), n_basis=5, order=3)

        coefficients = [[0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
                        [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12]]

        f = FDataBasis(bspline, coefficients)

        t = np.linspace(0, 1, 4)

        f_deriv = f.derivative()
        np.testing.assert_array_almost_equal(
            f_deriv(t).round(3),
            np.array([[2.927,  0.453, -1.229,  0.6],
                      [4.3, -1.599,  1.016, -2.52]])[..., np.newaxis]
        )

    def test_evaluation_grid_bspline(self):
        """Test the evaluation of FDataBasis with the grid option set to
            true. Nothing should be change due to the domain dimension is 1,
            but can accept the """
        bspline = BSpline(domain_range=(0, 1), n_basis=5, order=3)

        coefficients = [[0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
                        [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12]]

        f = FDataBasis(bspline, coefficients)
        t = np.linspace(0, 1, 4)

        res_test = f(t)

        # Different ways to pass the axes
        np.testing.assert_array_almost_equal(f(t, grid=True), res_test)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res_test)
        np.testing.assert_array_almost_equal(f([t], grid=True), res_test)
        np.testing.assert_array_almost_equal(
            f(np.atleast_2d(t), grid=True), res_test)

        # Number of axis different than the domain dimension (1)
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_composed_bspline(self):
        """Test the evaluation of FDataBasis the a matrix of times instead of
        a list of times """
        bspline = BSpline(domain_range=(0, 1), n_basis=5, order=3)

        coefficients = [[0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
                        [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12]]

        f = FDataBasis(bspline, coefficients)
        t = np.linspace(0, 1, 4)

        # Test same result than evaluation standart
        np.testing.assert_array_almost_equal(f([1]),
                                             f([[1], [1]],
                                               aligned=False))
        np.testing.assert_array_almost_equal(f(t), f(np.vstack((t, t)),
                                                     aligned=False))

        # Different evaluation times
        t_multiple = [[0, 0.5], [0.2, 0.7]]
        np.testing.assert_array_almost_equal(f(t_multiple[0])[0],
                                             f(t_multiple,
                                               aligned=False)[0])
        np.testing.assert_array_almost_equal(f(t_multiple[1])[1],
                                             f(t_multiple,
                                               aligned=False)[1])

    def test_domain_in_list_bspline(self):
        """Test the evaluation of FDataBasis"""

        for bspline in (BSpline(domain_range=[(0, 1)], n_basis=5, order=3),
                        BSpline(domain_range=((0, 1),), n_basis=5, order=3),
                        BSpline(domain_range=np.array((0, 1)), n_basis=5,
                                order=3),
                        BSpline(domain_range=np.array([(0, 1)]), n_basis=5,
                                order=3)
                        ):

            coefficients = [[0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
                            [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12]]

            f = FDataBasis(bspline, coefficients)

            t = np.linspace(0, 1, 4)

            res = np.array([[0.001, 0.564, 0.435, 0.33],
                            [0.018, 0.468, 0.371, 0.12]])[..., np.newaxis]

            np.testing.assert_array_almost_equal(f(t).round(3), res)
            np.testing.assert_array_almost_equal(f.evaluate(t).round(3), res)

        # Check error
        with np.testing.assert_raises(ValueError):
            BSpline(domain_range=[(0, 1), (0, 1)])


class TestBasisEvaluationMonomial(unittest.TestCase):

    def test_evaluation_simple_monomial(self):
        """Test the evaluation of FDataBasis"""

        monomial = Monomial(domain_range=(0, 2), n_basis=5)

        coefficients = np.array([[1,  2,  3,  4,  5],
                                 [6,  7,  8,  9, 10]])

        f = FDataBasis(monomial, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array(
            [[1.00, 1.56, 2.66, 4.79, 8.62, 15.00,
              25.00, 39.86, 61.03, 90.14, 129.00],
             [6.00, 7.81, 10.91, 16.32, 25.42, 40.00,
              62.21, 94.59, 140.08, 201.98, 284.00]])[..., np.newaxis]

        np.testing.assert_array_almost_equal(f(t).round(2), res)
        np.testing.assert_array_almost_equal(f.evaluate(t).round(2), res)

    def test_evaluation_point_monomial(self):
        """Test the evaluation of a single point FDataBasis"""
        monomial = Monomial(domain_range=(0, 1), n_basis=3)

        coefficients = [[1, 2, 3], [0.5, 1.4, 1.3]]

        f = FDataBasis(monomial, coefficients)

        # Test different ways of call f with a point
        res = np.array([[2.75], [1.525]])[..., np.newaxis]

        np.testing.assert_array_almost_equal(f([0.5]).round(4), res)
        np.testing.assert_array_almost_equal(f((0.5,)).round(4), res)
        np.testing.assert_array_almost_equal(f(0.5).round(4), res)
        np.testing.assert_array_almost_equal(f(np.array([0.5])).round(4), res)

        # Problematic case, should be accepted or no?
        #np.testing.assert_array_almost_equal(f(np.array(0.5)).round(4), res)

    def test_evaluation_derivative_monomial(self):
        """Test the evaluation of the derivative of a FDataBasis"""
        monomial = Monomial(domain_range=(0, 1), n_basis=3)

        coefficients = [[1, 2, 3], [0.5, 1.4, 1.3]]

        f = FDataBasis(monomial, coefficients)

        t = np.linspace(0, 1, 4)

        f_deriv = f.derivative()
        np.testing.assert_array_almost_equal(
            f_deriv(t).round(3),
            np.array([[2., 4., 6., 8.],
                      [1.4, 2.267, 3.133, 4.]])[..., np.newaxis]
        )

    def test_evaluation_grid_monomial(self):
        """Test the evaluation of FDataBasis with the grid option set to
            true. Nothing should be change due to the domain dimension is 1,
            but can accept the """
        monomial = Monomial(domain_range=(0, 1), n_basis=3)

        coefficients = [[1, 2, 3], [0.5, 1.4, 1.3]]

        f = FDataBasis(monomial, coefficients)
        t = np.linspace(0, 1, 4)

        res_test = f(t)

        # Different ways to pass the axes
        np.testing.assert_array_almost_equal(f(t, grid=True), res_test)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res_test)
        np.testing.assert_array_almost_equal(f([t], grid=True), res_test)
        np.testing.assert_array_almost_equal(
            f(np.atleast_2d(t), grid=True), res_test)

        # Number of axis different than the domain dimension (1)
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_composed_monomial(self):
        """Test the evaluation of FDataBasis the a matrix of times instead of
        a list of times """
        monomial = Monomial(domain_range=(0, 1), n_basis=3)

        coefficients = [[1, 2, 3], [0.5, 1.4, 1.3]]

        f = FDataBasis(monomial, coefficients)
        t = np.linspace(0, 1, 4)

        # Test same result than evaluation standart
        np.testing.assert_array_almost_equal(f([1]),
                                             f([[1], [1]],
                                               aligned=False))
        np.testing.assert_array_almost_equal(f(t), f(np.vstack((t, t)),
                                                     aligned=False))

        # Different evaluation times
        t_multiple = [[0, 0.5], [0.2, 0.7]]
        np.testing.assert_array_almost_equal(f(t_multiple[0])[0],
                                             f(t_multiple,
                                               aligned=False)[0])
        np.testing.assert_array_almost_equal(f(t_multiple[1])[1],
                                             f(t_multiple,
                                               aligned=False)[1])

    def test_domain_in_list_monomial(self):
        """Test the evaluation of FDataBasis"""

        for monomial in (Monomial(domain_range=[(0, 1)], n_basis=3),
                         Monomial(domain_range=((0, 1),), n_basis=3),
                         Monomial(domain_range=np.array((0, 1)), n_basis=3),
                         Monomial(domain_range=np.array([(0, 1)]), n_basis=3)):

            coefficients = [[1, 2, 3], [0.5, 1.4, 1.3]]

            f = FDataBasis(monomial, coefficients)

            t = np.linspace(0, 1, 4)

            res = np.array([[1., 2., 3.667, 6.],
                            [0.5, 1.111, 2.011, 3.2]])[..., np.newaxis]

            np.testing.assert_array_almost_equal(f(t).round(3), res)
            np.testing.assert_array_almost_equal(f.evaluate(t).round(3), res)


class TestBasisEvaluationVectorValued(unittest.TestCase):

    def test_vector_valued_constant(self):

        basis_first = Constant()
        basis_second = Constant()

        basis = VectorValued([basis_first, basis_second])

        fd = FDataBasis(basis=basis, coefficients=[[1, 2], [3, 4]])

        self.assertEqual(fd.dim_codomain, 2)

        res = np.array([[[1, 2]], [[3, 4]]])

        np.testing.assert_allclose(fd(0), res)

    def test_vector_valued_constant_monomial(self):

        basis_first = Constant(domain_range=(0, 5))
        basis_second = Monomial(n_basis=3, domain_range=(0, 5))

        basis = VectorValued([basis_first, basis_second])

        fd = FDataBasis(basis=basis, coefficients=[
                        [1, 2, 3, 4], [3, 4, 5, 6]])

        self.assertEqual(fd.dim_codomain, 2)

        np.testing.assert_allclose(fd.domain_range[0], (0, 5))

        res = np.array([[[1, 2], [1, 9], [1, 24]],
                        [[3, 4], [3, 15], [3, 38]]])

        np.testing.assert_allclose(fd([0, 1, 2]), res)


class TestBasisEvaluationTensor(unittest.TestCase):

    def test_tensor_monomial_constant(self):

        basis = Tensor([Monomial(n_basis=2), Constant()])

        fd = FDataBasis(basis=basis, coefficients=[1, 1])

        self.assertEqual(fd.dim_domain, 2)
        self.assertEqual(fd.dim_codomain, 1)

        np.testing.assert_allclose(fd([0., 0.]), [[[1.]]])

        np.testing.assert_allclose(fd([0.5, 0.5]), [[[1.5]]])

        np.testing.assert_allclose(
            fd([(0., 0.), (0.5, 0.5)]), [[[1.0], [1.5]]])

        fd_grid = fd.to_grid()

        fd2 = fd_grid.to_basis(basis)

        np.testing.assert_allclose(fd.coefficients, fd2.coefficients)


if __name__ == '__main__':
    print()
    unittest.main()




"""Implementation of Lp distances."""
from __future__ import annotations

import math
from typing import Optional, TypeVar, Union

import numpy as np
from typing_extensions import Final

from ...representation import FData
from ...representation._typing import NDArrayFloat
from ._lp_norms import LpNorm
from ._typing import Norm
from ._utils import NormInducedMetric, pairwise_metric_optimization

T = TypeVar("T", NDArrayFloat, FData)


class LpDistance(NormInducedMetric[Union[NDArrayFloat, FData]]):
    r"""
    Lp distance for functional data objects.

    Calculates the distance between two functional objects.

    For each pair of observations f and g the distance between them is defined
    as:

    .. math::
        d(x, y) = \| x - y \|_p

    where :math:`\| {}\cdot{} \|_p` denotes the :func:`Lp norm <lp_norm>`.

    The objects ``l1_distance``, ``l2_distance`` and ``linf_distance`` are
    instances of this class with commonly used values of ``p``, namely 1, 2 and
    infinity.

    Args:
        p: p of the lp norm. Must be greater or equal
            than 1. If ``p=math.inf`` it is used the L infinity metric.
            Defaults to 2.
        vector_norm: vector norm to apply. If it is a float, is the index of
            the multivariate lp norm. Defaults to the same as ``p``.

    Examples:
        Computes the distances between an object containing functional data
        corresponding to the functions y = 1 and y = x defined over the
        interval [0, 1] and another ones containing data of the functions y
        = 0 and y = x/2. The result then is an array 2x2 with the computed
        l2 distance between every pair of functions.

        >>> import skfda
        >>> import numpy as np
        >>>
        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([np.ones(len(x))], x)
        >>> fd2 =  skfda.FDataGrid([np.zeros(len(x))], x)
        >>>
        >>> distance = skfda.misc.metrics.LpDistance(p=2)
        >>> distance(fd, fd2).round(2)
        array([ 1.])


        If the functional data are defined over a different set of points of
        discretisation the functions returns an exception.

        >>> x = np.linspace(0, 2, 1001)
        >>> fd2 = skfda.FDataGrid([np.zeros(len(x)), x/2 + 0.5], x)
        >>> distance = skfda.misc.metrics.LpDistance(p=2)
        >>> distance(fd, fd2)
        Traceback (most recent call last):
            ...
        ValueError: ...

    """  # noqa: P102

    def __init__(
        self,
        p: float,
        vector_norm: Union[Norm[NDArrayFloat], float, None] = None,
    ) -> None:

        self.p = p
        self.vector_norm = vector_norm
        norm = LpNorm(p=p, vector_norm=vector_norm)

        super().__init__(norm)

    # This method is retyped here to work with either arrays or functions
    def __call__(self, elem1: T, elem2: T) -> NDArrayFloat:  # noqa: WPS612
        return super().__call__(elem1, elem2)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"p={self.p}, vector_norm={self.vector_norm})"
        )


l1_distance: Final = LpDistance(p=1)
l2_distance: Final = LpDistance(p=2)
linf_distance: Final = LpDistance(p=math.inf)


@pairwise_metric_optimization.register
def _pairwise_metric_optimization_lp_fdata(
    metric: LpDistance,
    elem1: FData,
    elem2: Optional[FData],
) -> NDArrayFloat:
    from ...misc import inner_product, inner_product_matrix

    vector_norm = metric.vector_norm

    if vector_norm is None:
        vector_norm = metric.p

    # Special case, the inner product is heavily optimized
    if metric.p == vector_norm == 2:
        diag1 = inner_product(elem1, elem1)
        diag2 = diag1 if elem2 is None else inner_product(elem2, elem2)

        if elem2 is None:
            elem2 = elem1

        inner_matrix = inner_product_matrix(elem1, elem2)

        distance_matrix_sqr = (
            -2 * inner_matrix
            + diag1[:, np.newaxis]
            + diag2[np.newaxis, :]
        )

        np.clip(
            distance_matrix_sqr,
            a_min=0,
            a_max=None,
            out=distance_matrix_sqr,
        )

        return np.sqrt(distance_matrix_sqr)

    return NotImplemented


def lp_distance(
    fdata1: T,
    fdata2: T,
    *,
    p: float,
    vector_norm: Union[Norm[NDArrayFloat], float, None] = None,
) -> NDArrayFloat:
    r"""
    Lp distance for FDataGrid objects.

    Calculates the distance between two functional objects.

    For each pair of observations f and g the distance between them is defined
    as:

    .. math::
        d(f, g) = d(g, f) = \| f - g \|_p

    where :math:`\| {}\cdot{} \|_p` denotes the :func:`Lp norm <lp_norm>`.

    Note:
        This function is a wrapper of :class:`LpDistance`, available only for
        convenience. As the parameter ``p`` is mandatory, it cannot be used
        where a fully-defined metric is required: use an instance of
        :class:`LpDistance` in those cases.

    Args:
        fdata1: First FData object.
        fdata2: Second FData object.
        p: p of the lp norm. Must be greater or equal
            than 1. If ``p=math.inf`` it is used the L infinity metric.
            Defaults to 2.
        vector_norm: vector norm to apply. If it is a float, is the index of
            the multivariate lp norm. Defaults to the same as ``p``.

    Returns:
        Numpy vector where the i-th coordinate has the distance between the
        i-th element of the first object and the i-th element of the second
        one.

    Examples:
        Computes the distances between an object containing functional data
        corresponding to the functions y = 1 and y = x defined over the
        interval [0, 1] and another ones containing data of the functions y
        = 0 and y = x/2. The result then is an array of size 2 with the
        computed l2 distance between the functions in the same position in
        both.

        >>> import skfda
        >>> import numpy as np
        >>>
        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([np.ones(len(x)), x], x)
        >>> fd2 =  skfda.FDataGrid([np.zeros(len(x)), x/2], x)
        >>>
        >>> skfda.misc.metrics.lp_distance(fd, fd2, p=2).round(2)
        array([ 1.  ,  0.29])

        If the functional data are defined over a different set of points of
        discretisation the functions returns an exception.

        >>> x = np.linspace(0, 2, 1001)
        >>> fd2 = skfda.FDataGrid([np.zeros(len(x)), x/2 + 0.5], x)
        >>> skfda.misc.metrics.lp_distance(fd, fd2, p=2)
        Traceback (most recent call last):
            ...
        ValueError: ...

    See also:
        :class:`~skfda.misc.metrics.LpDistance`

    """  # noqa: P102
    return LpDistance(p=p, vector_norm=vector_norm)(fdata1, fdata2)



"""Dim reduction."""
from __future__ import annotations

import importlib
from typing import Any

from . import feature_extraction, variable_selection


def __getattr__(name: str) -> Any:
    if name == "projection":
        return importlib.import_module(f".{name}", __name__)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from ._boxplot import BoxplotOutlierDetector
from ._directional_outlyingness import (
    MSPlotOutlierDetector,
    directional_outlyingness_stats,
)
from ._outliergram import OutliergramOutlierDetector
from .neighbors_outlier import LocalOutlierFactor



from skfda import FDataGrid
from skfda.exploratory.depth import IntegratedDepth
from skfda.exploratory.visualization import Boxplot, SurfaceBoxplot
import unittest

import numpy as np


class TestBoxplot(unittest.TestCase):

    def test_fdboxplot_univariate(self):
        data_matrix = [[1, 1, 2, 3, 2.5, 2],
                       [0.5, 0.5, 1, 2, 1.5, 1],
                       [-1, -1, -0.5, 1, 1, 0.5],
                       [-0.5, -0.5, -0.5, -1, -1, -1]]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        fdataBoxplot = Boxplot(fd, depth_method=IntegratedDepth())
        np.testing.assert_array_equal(
            fdataBoxplot.median.ravel(),
            np.array([-1., -1., -0.5, 1., 1., 0.5]))
        np.testing.assert_array_equal(
            fdataBoxplot.central_envelope[0].ravel(),
            np.array([-1., -1., -0.5, -1., -1., -1.]))
        np.testing.assert_array_equal(
            fdataBoxplot.central_envelope[1].ravel(),
            np.array([-0.5, -0.5, -0.5, 1., 1., 0.5]))
        np.testing.assert_array_equal(
            fdataBoxplot.non_outlying_envelope[0].ravel(),
            np.array([-1., -1., -0.5, -1., -1., -1.]))
        np.testing.assert_array_equal(
            fdataBoxplot.non_outlying_envelope[1].ravel(),
            np.array([-0.5, -0.5, -0.5,  1.,  1.,  0.5]))
        self.assertEqual(len(fdataBoxplot.envelopes), 1)
        np.testing.assert_array_equal(
            fdataBoxplot.envelopes[0],
            fdataBoxplot.central_envelope)
        np.testing.assert_array_equal(fdataBoxplot.outliers,
                                      np.array([True, True, False, False]))


if __name__ == '__main__':
    print()
    unittest.main()




from skfda import FDataGrid
from skfda.representation.interpolation import SplineInterpolation
import unittest

import numpy as np


# TODO: Unitest for grids with domain dimension > 1
class TestEvaluationSpline1_1(unittest.TestCase):
    """Test the evaluation of a grid spline interpolation with
    domain and image dimension equal to 1.
    """

    def setUp(self):
        # Data matrix of a datagrid with a dimension of domain and image equal
        # to 1.

        # Matrix of functions (x**2, (9-x)**2)
        self.data_matrix_1_1 = [np.arange(10)**2,
                                np.arange(start=9, stop=-1, step=-1)**2]

    def test_evaluation_linear_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(
            f(np.arange(10))[..., 0], self.data_matrix_1_1)

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(
            f([0.5, 1.5, 2.5]),
            np.array([[[0.5],  [2.5],  [6.5]],
                      [[72.5], [56.5], [42.5]]]))

    def test_evaluation_linear_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10))

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3).round(1),
                                             np.array([[[28.3]], [[13.9]]]))
        np.testing.assert_array_almost_equal(
            f([3]), np.array([[[9.]], [[36.]]]))
        np.testing.assert_array_almost_equal(
            f((2,)), np.array([[[4.]], [[49.]]]))

    def test_evaluation_linear_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(np.arange(10))[..., 0],
                                             self.data_matrix_1_1)

        res = np.array([[[0.5],  [2.5],  [6.5]], [[72.5], [56.5], [42.5]]])
        t = [0.5, 1.5, 2.5]

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res)
        np.testing.assert_array_almost_equal(f([t], grid=True), res)
        # Single point with grid
        np.testing.assert_array_almost_equal(f(3, grid=True),
                                             np.array([[[9.]], [[36.]]]))

        # Check erroneous axis
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_linear_composed(self):

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10))

        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(f([[1], [8]],
                                               aligned=False),
                                             np.array([[[1.]], [[1.]]]))

        t = np.linspace(4, 6, 4)
        np.testing.assert_array_almost_equal(
            f([t, 9 - t], aligned=False).round(2),
            np.array([[[16.], [22.], [28.67], [36.]],
                      [[16.], [22.], [28.67], [36.]]]))

        # Same length than nsample
        t = np.linspace(4, 6, 2)
        np.testing.assert_array_almost_equal(
            f([t, 9 - t], aligned=False).round(2),
            np.array([[[16.], [36.]], [[16.], [36.]]]))

    def test_evaluation_cubic_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                      interpolation=SplineInterpolation(3))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(np.arange(10)).round(1)[..., 0],
                                             self.data_matrix_1_1)

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(
            f([0.5, 1.5, 2.5]).round(2),
            np.array([[[0.25],  [2.25],  [6.25]],
                      [[72.25], [56.25], [42.25]]]))

    def test_evaluation_cubic_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                      interpolation=SplineInterpolation(3))

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3).round(3),
                                             np.array([[[28.09]], [[13.69]]]))

        np.testing.assert_array_almost_equal(
            f([3]).round(3), np.array([[[9.]], [[36.]]]))
        np.testing.assert_array_almost_equal(
            f((2,)).round(3), np.array([[[4.]], [[49.]]]))

    def test_evaluation_cubic_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                      interpolation=SplineInterpolation(3))

        t = [0.5, 1.5, 2.5]
        res = np.array([[[0.25],  [2.25],  [6.25]],
                        [[72.25], [56.25], [42.25]]])

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True).round(3), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True).round(3), res)
        np.testing.assert_array_almost_equal(f([t], grid=True).round(3), res)
        # Single point with grid
        np.testing.assert_array_almost_equal(
            f(3, grid=True), np.array([[[9.]], [[36.]]]))

        # Check erroneous axis
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_cubic_composed(self):

        f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                      interpolation=SplineInterpolation(3))

        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(
            f([[1], [8]], aligned=False).round(3),
            np.array([[[1.]], [[1.]]]))

        t = np.linspace(4, 6, 4)
        np.testing.assert_array_almost_equal(
            f([t, 9 - t], aligned=False).round(2),
            np.array([[[16.], [21.78], [28.44], [36.]],
                      [[16.], [21.78], [28.44], [36.]]]))

        # Same length than nsample
        t = np.linspace(4, 6, 2)
        np.testing.assert_array_almost_equal(
            f([t, 9 - t], aligned=False).round(3),
            np.array([[[16.], [36.]], [[16.], [36.]]]))

    def test_evaluation_nodes(self):
        """Test interpolation in nodes for all dimensions"""

        for degree in range(1, 6):
            interpolation = SplineInterpolation(degree)

            f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                          interpolation=interpolation)

            # Test interpolation in nodes
            np.testing.assert_array_almost_equal(
                f(np.arange(10)).round(5)[..., 0],
                self.data_matrix_1_1)

    def test_error_degree(self):

        with np.testing.assert_raises(ValueError):
            interpolation = SplineInterpolation(7)
            f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                          interpolation=interpolation)
            f(1)

        with np.testing.assert_raises(ValueError):
            interpolation = SplineInterpolation(0)
            f = FDataGrid(self.data_matrix_1_1, grid_points=np.arange(10),
                          interpolation=interpolation)
            f(1)


class TestEvaluationSpline1_n(unittest.TestCase):
    """Test the evaluation of a grid spline interpolation with
    domain dimension equal to 1 and arbitary image dimension.
    """

    def setUp(self):
        # Data matrix of a datagrid with a dimension of domain and image equal
        # to 1.

        # Matrix of functions (x**2, (9-x)**2)

        self.t = np.arange(10)

        data_1 = np.array([np.arange(10)**2,
                           np.arange(start=9, stop=-1, step=-1)**2])
        data_2 = np.sin(np.pi / 81 * data_1)

        self.data_matrix_1_n = np.dstack((data_1, data_2))

        self.interpolation = SplineInterpolation(interpolation_order=2)

    def test_evaluation_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_n, grid_points=np.arange(10),
                      interpolation=self.interpolation)

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(self.t), self.data_matrix_1_n)
        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f([1.5, 2.5, 3.5]),
                                             np.array([[[2.25,  0.087212],
                                                        [6.25,  0.240202],
                                                        [12.25,  0.45773]],
                                                       [[56.25,  0.816142],
                                                        [42.25,  0.997589],
                                                        [30.25,  0.922146]]]
                                                      )
                                             )

    def test_evaluation_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_n, grid_points=np.arange(10),
                      interpolation=self.interpolation)

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3),
                                             np.array([[[28.09,  0.885526]],
                                                       [[13.69,  0.50697]]]
                                                      )
                                             )

    def test_evaluation_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_n, grid_points=np.arange(10),
                      interpolation=SplineInterpolation(2))

        t = [1.5, 2.5, 3.5]
        res = np.array([[[2.25,  0.08721158],
                         [6.25,  0.24020233],
                         [12.25,  0.4577302]],
                        [[56.25,  0.81614206],
                         [42.25,  0.99758925],
                         [30.25,  0.92214607]]])

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res)
        np.testing.assert_array_almost_equal(f([t], grid=True), res)

        # Check erroneous axis
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_composed(self):

        f = FDataGrid(self.data_matrix_1_n, grid_points=self.t,
                      interpolation=self.interpolation)

        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(f([[1], [4]],
                                               aligned=False)[0],
                                             f(1)[0])
        np.testing.assert_array_almost_equal(f([[1], [4]],
                                               aligned=False)[1],
                                             f(4)[1])

    def test_evaluation_nodes(self):
        """Test interpolation in nodes for all dimensions"""

        for degree in range(1, 6):
            interpolation = SplineInterpolation(degree)

            f = FDataGrid(self.data_matrix_1_n, grid_points=np.arange(10),
                          interpolation=interpolation)

            # Test interpolation in nodes
            np.testing.assert_array_almost_equal(f(np.arange(10)),
                                                 self.data_matrix_1_n)


class TestEvaluationSplinem_n(unittest.TestCase):
    """Test the evaluation of a grid spline interpolation with
    arbitrary domain dimension and arbitary image dimension.
    """

    def test_evaluation_center_and_extreme_points_linear(self):
        """Test linear interpolation in the middle point of a grid square."""

        dim_codomain = 4
        n_samples = 2

        @np.vectorize
        def coordinate_function(*args):
            _, *domain_indexes, _ = args
            return np.sum(domain_indexes)

        for dim_domain in range(1, 6):
            grid_points = [np.array([0, 1]) for _ in range(dim_domain)]
            data_matrix = np.fromfunction(
                function=coordinate_function,
                shape=(n_samples,) + (2,) * dim_domain + (dim_codomain,))

            f = FDataGrid(data_matrix, grid_points=grid_points)

            evaluation = f([[0.] * dim_domain, [0.5] *
                            dim_domain, [1.] * dim_domain])

            self.assertEqual(evaluation.shape, (n_samples, 3, dim_codomain))

            for i in range(n_samples):
                for j in range(dim_codomain):
                    np.testing.assert_array_almost_equal(
                        evaluation[i, ..., j],
                        [0, dim_domain * 0.5, dim_domain])


if __name__ == '__main__':
    print()
    unittest.main()



"""Functional data descriptive statistics."""

from builtins import isinstance
from typing import Optional, TypeVar, Union

import numpy as np
from scipy import integrate
from scipy.stats import rankdata

from ...misc.metrics import Metric, l2_distance, l2_norm
from ...representation import FData, FDataGrid
from ..depth import Depth, ModifiedBandDepth

F = TypeVar('F', bound=FData)


def mean(X: F) -> F:
    """Compute the mean of all the samples in a FData object.

    Args:
        X: Object containing all the samples whose mean is wanted.


    Returns:
        A :term:`functional data object` with just one sample representing
        the mean of all the samples in the original object.

    """
    return X.mean()


def var(X: FData) -> FDataGrid:
    """Compute the variance of a set of samples in a FDataGrid object.

    Args:
        X: Object containing all the set of samples whose variance is desired.

    Returns:
        A :term:`functional data object` with just one sample representing the
        variance of all the samples in the original object.

    """
    return X.var()


def gmean(X: FDataGrid) -> FDataGrid:
    """Compute the geometric mean of all the samples in a FDataGrid object.

    Args:
        X: Object containing all the samples whose geometric mean is wanted.

    Returns:
        A :term:`functional data object` with just one sample representing the
        geometric mean of all the samples in the original object.

    """
    return X.gmean()


def cov(X: FData) -> FDataGrid:
    """Compute the covariance.

    Calculates the covariance matrix representing the covariance of the
    functional samples at the observation points.

    Args:
        X: Object containing different samples of a functional variable.

    Returns:
        A :term:`functional data object` with just one sample representing the
        covariance of all the samples in the original object.

    """
    return X.cov()


def modified_epigraph_index(X: FDataGrid) -> np.ndarray:
    """
    Calculate the Modified Epigraph Index of a FDataGrid.

    The MEI represents the mean time a curve stays below other curve.
    In this case we will calculate the MEI for each curve in relation
    with all the other curves of our dataset.

    """
    interval_len = (
        X.domain_range[0][1]
        - X.domain_range[0][0]
    )

    # Array containing at each point the number of curves
    # are above it.
    num_functions_above = rankdata(
        -X.data_matrix,
        method='max',
        axis=0,
    ) - 1

    integrand = num_functions_above

    for d, s in zip(X.domain_range, X.grid_points):
        integrand = integrate.simps(
            integrand,
            x=s,
            axis=1,
        )
        interval_len = d[1] - d[0]
        integrand /= interval_len

    integrand /= X.n_samples

    return integrand.flatten()


def depth_based_median(
    X: FDataGrid,
    depth_method: Optional[Depth] = None,
) -> FDataGrid:
    """Compute the median based on a depth measure.

    The depth based median is the deepest curve given a certain
    depth measure.

    Args:
        X: Object containing different samples of a
            functional variable.
        depth_method: Method used to order the data. Defaults to
            :func:`modified band
            depth <skfda.exploratory.depth.ModifiedBandDepth>`.

    Returns:
        Object containing the computed depth_based median.

    See also:
        :func:`geometric_median`

    """
    if depth_method is None:
        depth_method = ModifiedBandDepth()

    depth = depth_method(X)
    indices_descending_depth = (-depth).argsort(axis=0)

    # The median is the deepest curve
    return X[indices_descending_depth[0]]


T = TypeVar('T', bound=Union[np.ndarray, FData])


def _weighted_average(X: T, weights: np.ndarray) -> T:

    return (
        (X * weights).sum() if isinstance(X, FData)
        # To support also multivariate data
        else (X.T * weights).T.sum(axis=0)
    )


def geometric_median(
    X: T,
    *,
    tol: float = 1.e-8,
    metric: Metric[T] = l2_distance,
) -> T:
    r"""Compute the geometric median.

    The sample geometric median is the point that minimizes the :math:`L_1`
    norm of the vector of distances to all observations:

    .. math::

        \underset{y \in L(\mathcal{T})}{\arg \min}
        \sum_{i=1}^N \left \| x_i-y \right \|

    It uses the corrected Weiszfeld algorithm to compute the median.

    Args:
        X: Object containing different samples of a
            functional variable.
        tol: tolerance used to check convergence.
        metric: metric used to compute the vector of distances. By
            default is the :math:`L_2` distance.

    Returns:
        Object containing the computed geometric median.

    Example:

        >>> from skfda import FDataGrid
        >>> data_matrix = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
        >>> X = FDataGrid(data_matrix)
        >>> median = geometric_median(X)
        >>> median.data_matrix[0, ..., 0]
        array([ 1. ,  1. ,  3. ,  0.5])

    See also:
        :func:`depth_based_median`

    References:
        .. footbibliography::

            gervini_2008_estimation

    """
    weights = np.full(len(X), 1 / len(X))
    median = _weighted_average(X, weights)
    distances = metric(X, median)

    while True:
        zero_distances = (distances == 0)
        n_zeros = np.sum(zero_distances)
        weights_new = (
            (1 / distances) / np.sum(1 / distances) if n_zeros == 0
            else (1 / n_zeros) * zero_distances
        )

        median_new = _weighted_average(X, weights_new)

        if l2_norm(median_new - median) < tol:
            return median_new

        distances = metric(X, median_new)

        weights, median = (weights_new, median_new)


def trim_mean(
    X: F,
    proportiontocut: float,
    *,
    depth_method: Optional[Depth[F]] = None,
) -> FDataGrid:
    """Compute the trimmed means based on a depth measure.

    The trimmed means consists in computing the mean function without a
    percentage of least deep curves. That is, we first remove the least deep
    curves and then we compute the mean as usual.

    Note that in scipy the leftmost and rightmost proportiontocut data are
    removed. In this case, as we order the data by the depth, we only remove
    those that have the least depth values.

    Args:
        X: Object containing different samples of a
            functional variable.
        proportiontocut: Indicates the percentage of functions to
            remove. It is not easy to determine as it varies from dataset to
            dataset.
        depth_method: Method used to order the data. Defaults to
            :func:`modified band depth
            <skfda.exploratory.depth.ModifiedBandDepth>`.

    Returns:
        Object containing the computed trimmed mean.

    """
    if depth_method is None:
        depth_method = ModifiedBandDepth()

    n_samples_to_keep = (len(X) - int(len(X) * proportiontocut))

    # compute the depth of each curve and store the indexes in descending order
    depth = depth_method(X)
    indices_descending_depth = (-depth).argsort(axis=0)

    trimmed_curves = X[indices_descending_depth[:n_samples_to_keep]]

    return trimmed_curves.mean()
